import json
import os
import re
import sys
from pathlib import Path

from crytic_compile import CryticCompile, save_to_zip
from crytic_compile.utils.zip import load_from_zip
from deepdiff import DeepDiff
from packaging.version import parse as parse_version
from solc_select.solc_select import install_artifacts as install_solc_versions
from solc_select.solc_select import installed_versions as get_installed_solc_versions

from slither import Slither
from slither.printers.guidance.echidna import Echidna

E2E_ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
TEST_ROOT = os.path.join(E2E_ROOT, "solc_parsing", "test_data")


class Test:
    def __init__(
        self,
        test_file: str,
        solc_versions: list[str],
        disable_legacy: bool = False,
        solc_args: str | None = None,
    ):
        self.solc_versions = solc_versions
        self.test_file = test_file
        self.disable_legacy = disable_legacy

        versions_with_flavors: list[tuple[str, str, str]] = []
        flavors = ["compact"]
        if not self.disable_legacy:
            flavors += ["legacy"]
        for version in solc_versions:
            for flavor in flavors:
                # No legacy AST format for >0.8
                legacy_unavailable = flavor == "legacy" and parse_version(version) >= parse_version(
                    "0.8"
                )
                # No compact AST format for <0.4.12
                compact_unavailable = flavor == "compact" and parse_version(
                    version
                ) < parse_version("0.4.12")
                if legacy_unavailable or compact_unavailable:
                    continue
                versions_with_flavors.append((version, flavor, solc_args))
        self.versions_with_flavors = versions_with_flavors


def generate_output(sl: Slither) -> dict[str, dict[str, str]]:
    output = {}
    for contract in sl.contracts:
        output[contract.name] = {}

        for func_or_modifier in contract.functions + contract.modifiers:
            output[contract.name][func_or_modifier.full_name] = (
                func_or_modifier.slithir_cfg_to_dot_str(skip_expressions=True)
            )

    return output


def make_version(minor: int, patch_min: int, patch_max: int) -> list[str]:
    return [f"0.{minor}.{x}" for x in range(patch_min, patch_max + 1)]


VERSIONS_04 = make_version(4, 0, 26)
VERSIONS_05 = make_version(5, 0, 17)
VERSIONS_06 = make_version(6, 0, 12)
VERSIONS_07 = make_version(7, 0, 6)
VERSIONS_08 = make_version(8, 0, 15)

ALL_VERSIONS = VERSIONS_04 + VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08

ALL_TESTS = [
    Test("using-for-0.4.0.sol", ["0.4.0"]),
    Test(
        "using-for-0.4.1.sol",
        make_version(4, 1, 26) + VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
    ),
    Test(
        "top-level-import-0.4.0.sol",
        VERSIONS_04 + VERSIONS_05 + VERSIONS_06 + ["0.7.0"],
    ),
    Test("top-level-import-0.7.1.sol", make_version(7, 1, 6) + VERSIONS_08),
    Test("function-0.4.0.sol", make_version(4, 0, 15)),
    Test("function-0.4.16.sol", make_version(4, 16, 21)),
    Test("function-0.4.22.sol", ["0.4.22"]),
    Test("function-0.4.23.sol", make_version(4, 23, 26)),
    Test("function-0.5.0.sol", VERSIONS_05),
    # TODO: legacy is failing
    Test("function-0.6.0.sol", VERSIONS_06, disable_legacy=True),
    # TODO: failing
    # Test("function-0.7.0.sol", ["0.7.0"]),
    # TODO: legacy failing with 0.7
    Test("function-0.7.1.sol", make_version(7, 1, 6), disable_legacy=True),
    Test("function-0.7.1.sol", VERSIONS_08),
    Test(
        "top-level-import-bis-0.4.0.sol",
        VERSIONS_04 + VERSIONS_05 + VERSIONS_06 + ["0.7.0"],
    ),
    Test("top-level-import-bis-0.7.1.sol", make_version(7, 1, 6) + VERSIONS_08),
    # TODO: failing
    # Test(
    #     "variabledeclaration-0.4.0.sol",
    #     make_version(4, 0, 23),
    #     disable_legacy=True,
    # ),
    # # TODO: failing
    # Test(
    #     "variabledeclaration-0.4.24.sol",
    #     make_version(4, 24, 26),
    #     disable_legacy=True,
    # ),
    Test(
        "variabledeclaration-0.5.0.sol",
        VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
        disable_legacy=True,
    ),
    Test("functioncall-0.4.0.sol", VERSIONS_04),
    Test(
        "functioncall-0.4.5.sol",
        make_version(4, 5, 9),
    ),
    Test(
        "functioncall-0.5.0.sol",
        make_version(5, 0, 2),
    ),
    Test(
        "functioncall-0.5.3.sol",
        make_version(5, 3, 17),
    ),
    Test(
        "functioncall-0.6.0.sol",
        make_version(6, 0, 1),
    ),
    Test(
        "functioncall-0.6.2.sol",
        make_version(6, 2, 7),
    ),
    Test(
        "functioncall-0.6.8.sol",
        make_version(6, 8, 12),
    ),
    Test(
        "functioncall-0.7.0.sol",
        VERSIONS_07,
    ),
    Test(
        "functioncall-0.8.0.sol",
        VERSIONS_08,
    ),
    Test(
        "break-all.sol",
        ALL_VERSIONS,
    ),
    Test(
        "top-level-nested-import-0.4.0.sol",
        VERSIONS_04 + VERSIONS_05 + VERSIONS_06 + ["0.7.0"],
    ),
    Test(
        "top-level-nested-import-0.7.1.sol",
        make_version(7, 1, 6) + VERSIONS_08,
    ),
    # + [f"variable_0.6.{ver}_legacy" for ver in range(5, 23)]
    # + [f"variable_0.6.{ver}_compact" for ver in range(5, 23)]
    # + [f"variable_0.7.{ver}_legacy" for ver in range(0, 2)]
    # + [f"variable_0.7.{ver}_compact" for ver in range(0, 2)]
    Test(
        "call_to_variable-all.sol",
        ALL_VERSIONS,
    ),
    Test("yul-0.4.0.sol", ["0.4.0"]),
    Test("yul-0.4.1.sol", make_version(4, 1, 10)),
    Test(
        "yul-0.4.11.sol",
        make_version(4, 11, 26) + VERSIONS_05 + VERSIONS_06,
    ),
    Test("yul-0.7.0.sol", make_version(7, 0, 4)),
    Test("yul-0.7.5.sol", make_version(7, 5, 6)),
    Test("yul-0.8.0.sol", ["0.8.33"]),
    Test("pragma-0.4.0.sol", VERSIONS_04),
    Test("pragma-0.5.0.sol", VERSIONS_05),
    Test("pragma-0.6.0.sol", VERSIONS_06),
    Test("pragma-0.7.0.sol", VERSIONS_07),
    Test("pragma-0.8.0.sol", VERSIONS_08),
    Test(
        "assembly-all.sol",
        ALL_VERSIONS,
    ),
    Test("struct-0.4.0.sol", VERSIONS_04 + VERSIONS_05),
    # TODO: legacy failing
    Test(
        "struct-0.6.0.sol",
        VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
        disable_legacy=True,
    ),
    # TODO: currently failing
    # Test("emit-0.4.0.sol", VERSIONS_04),
    # Test(
    #     "emit-0.4.8.sol",
    #     make_version(4, 8, 20)
    # ),
    # Test(
    #     "emit-0.4.21.sol",
    #     make_version(4, 21, 26)
    # ),
    Test(
        "emit-0.5.0.sol",
        VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
    ),
    # TODO: failing
    # Test("import-0.4.0.sol", VERSIONS_04),
    # Test(
    #     "import-0.4.3.sol",
    #     make_version(4, 3, 9) + VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
    # ),
    Test("tupleexpression-0.4.0.sol", make_version(4, 0, 23)),
    Test("tupleexpression-0.4.24.sol", make_version(4, 24, 26) + VERSIONS_05),
    Test(
        "tupleexpression-0.5.3.sol",
        make_version(5, 3, 9) + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
    ),
    Test("literal-0.4.0.sol", make_version(4, 0, 9)),
    Test("literal-0.4.10.sol", make_version(4, 10, 26)),
    Test("literal-0.5.0.sol", VERSIONS_05),
    Test("literal-0.6.0.sol", VERSIONS_06),
    # TODO: failing
    # Test("literal-0.7.0.sol", VERSIONS_07 + VERSIONS_08),
    Test("memberaccess-0.4.0.sol", VERSIONS_04 + VERSIONS_05),
    Test(
        "memberaccess-0.5.3.sol",
        make_version(5, 3, 9),
    ),
    # TODO:  Legacy failing from 0.6
    Test(
        "memberaccess-0.5.3.sol",
        VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
        disable_legacy=True,
    ),
    Test(
        "memberaccess-0.6.7.sol",
        ["0.6.7"],
        disable_legacy=True,
    ),
    Test(
        "memberaccess-0.6.8.sol",
        ["0.6.8"] + VERSIONS_07 + VERSIONS_08,
        disable_legacy=True,
    ),
    Test("throw-0.4.0.sol", VERSIONS_04),
    Test(
        "throw-0.5.0.sol",
        VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
    ),
    Test(
        "top_level_variable2-0.4.0.sol",
        VERSIONS_04 + VERSIONS_05 + VERSIONS_06 + VERSIONS_07,
    ),
    Test("top_level_variable2-0.8.0.sol", VERSIONS_08),
    Test(
        "comment-all.sol",
        ALL_VERSIONS,
    ),
    Test("assignment-0.4.0.sol", VERSIONS_04),
    Test(
        "assignment-0.4.7.sol",
        make_version(4, 7, 9) + VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
    ),
    Test(
        "event-all.sol",
        ALL_VERSIONS,
    ),
    # TODO: legacy not working
    Test(
        "indexrangeaccess-0.4.0.sol",
        VERSIONS_04 + VERSIONS_05 + ["0.6.0"],
    ),
    Test("indexrangeaccess-0.4.0.sol", ["0.6.0"], disable_legacy=True),
    Test(
        "indexrangeaccess-0.6.1.sol",
        make_version(6, 1, 12) + VERSIONS_07 + VERSIONS_08,
        disable_legacy=True,
    ),
    Test("variable-0.4.0.sol", VERSIONS_04),
    Test("variable-0.4.5.sol", make_version(4, 5, 13)),
    Test("variable-0.4.14.sol", make_version(4, 14, 15)),
    Test("variable-0.4.16.sol", make_version(4, 16, 26)),
    Test("variable-0.5.0.sol", VERSIONS_05 + make_version(6, 0, 4)),
    Test("variable-0.6.5.sol", make_version(6, 5, 8)),
    Test("variable-0.6.9.sol", make_version(6, 9, 12) + VERSIONS_07),
    Test("variable-0.8.0.sol", VERSIONS_08),
    Test(
        "continue-all.sol",
        ALL_VERSIONS,
    ),
    Test(
        "if-all.sol",
        ALL_VERSIONS,
    ),
    Test(
        "modifier-all.sol",
        VERSIONS_04 + VERSIONS_05 + VERSIONS_06,
    ),
    Test(
        "modifier-0.7.0.sol",
        VERSIONS_07 + VERSIONS_08,
    ),
    Test("library_implicit_conversion-0.4.0.sol", VERSIONS_04),
    Test(
        "library_implicit_conversion-0.5.0.sol",
        VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
    ),
    Test("units_and_global_variables-0.4.0.sol", VERSIONS_04),
    Test("units_and_global_variables-0.5.0.sol", make_version(5, 0, 3)),
    Test("units_and_global_variables-0.5.4.sol", make_version(5, 4, 17)),
    Test("units_and_global_variables-0.6.0.sol", VERSIONS_06),
    Test("units_and_global_variables-0.7.0.sol", VERSIONS_07),
    Test("units_and_global_variables-0.8.0.sol", VERSIONS_08),
    Test("units_and_global_variables-0.8.4.sol", make_version(8, 4, 6)),
    Test("units_and_global_variables-0.8.7.sol", make_version(8, 7, 9)),
    Test("global_variables-0.8.18.sol", make_version(8, 18, 18)),
    Test(
        "push-all.sol",
        ALL_VERSIONS,
    ),
    Test(
        "indexaccess-all.sol",
        ALL_VERSIONS,
    ),
    Test("minmax-0.4.0.sol", VERSIONS_04 + VERSIONS_05 + VERSIONS_06),
    Test(
        "minmax-0.6.8.sol",
        make_version(6, 8, 9) + VERSIONS_07 + VERSIONS_08,
    ),
    Test(
        "minmax-0.8.8.sol",
        make_version(8, 8, 15),
    ),
    Test("dowhile-0.4.0.sol", VERSIONS_04),
    Test(
        "dowhile-0.4.5.sol",
        make_version(4, 5, 9) + VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
    ),
    Test(
        "custom_error-0.4.0.sol",
        ALL_VERSIONS,
    ),
    Test("custom_error-0.8.4.sol", make_version(8, 4, 15)),
    Test("custom-error-selector.sol", make_version(8, 4, 15)),
    Test(
        "top-level-0.4.0.sol",
        VERSIONS_04 + VERSIONS_05 + VERSIONS_06 + ["0.7.0"],
    ),
    Test("top-level-0.7.1.sol", make_version(7, 1, 3)),
    Test("top-level-0.7.4.sol", make_version(7, 4, 6) + VERSIONS_08),
    Test("contract-0.4.0.sol", make_version(4, 0, 21)),
    Test("contract-0.4.22.sol", make_version(4, 22, 26) + VERSIONS_05),
    Test("contract-0.6.0.sol", VERSIONS_06 + VERSIONS_07 + VERSIONS_08),
    Test(
        "import_interface_with_struct_from_top_level-0.4.0.sol",
        VERSIONS_04 + VERSIONS_05 + VERSIONS_06 + make_version(7, 0, 5),
    ),
    Test(
        "import_interface_with_struct_from_top_level-0.7.6.sol",
        ["0.7.6"] + VERSIONS_08,
    ),
    Test("scope-0.4.0.sol", VERSIONS_04),
    Test(
        "scope-0.5.0.sol",
        VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
    ),
    Test(
        "conditional-all.sol",
        ALL_VERSIONS,
    ),
    Test(
        "for-all.sol",
        ALL_VERSIONS,
    ),
    Test("trycatch-0.4.0.sol", VERSIONS_04 + VERSIONS_05),
    # TODO: legacy failing
    Test(
        "trycatch-0.6.0.sol",
        VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
        disable_legacy=True,
    ),
    Test(
        "unchecked-0.4.0.sol",
        VERSIONS_04 + VERSIONS_05 + VERSIONS_06 + VERSIONS_07,
    ),
    Test("unchecked-0.8.0.sol", VERSIONS_08),
    Test(
        "return-all.sol",
        ALL_VERSIONS,
    ),
    Test("binaryoperation-0.4.0.sol", VERSIONS_04),
    Test(
        "binaryoperation-0.4.7.sol",
        make_version(4, 7, 9) + VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + make_version(8, 0, 12),
    ),
    Test("newexpression-0.4.0.sol", VERSIONS_04),
    Test(
        "newexpression-0.5.0.sol",
        VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
    ),
    Test(
        "enum-0.4.0.sol",
        VERSIONS_04 + VERSIONS_05 + VERSIONS_06 + VERSIONS_07,
    ),
    Test("enum-0.8.0.sol", VERSIONS_08),
    Test(
        "top_level_variable-0.4.0.sol",
        VERSIONS_04 + VERSIONS_05 + VERSIONS_06 + VERSIONS_07,
    ),
    Test("top_level_variable-0.8.0.sol", VERSIONS_08),
    Test("unaryexpression-0.4.0.sol", VERSIONS_04),
    Test(
        "unaryexpression-0.5.0.sol",
        VERSIONS_05 + VERSIONS_06 + VERSIONS_07 + VERSIONS_08,
    ),
    Test(
        "while-all.sol",
        ALL_VERSIONS,
    ),
    Test(
        "complex_imports/import_free/Caller.sol",
        ["0.8.2"],
    ),
    Test("custom_error_with_state_variable.sol", make_version(8, 4, 12)),
    Test("complex_imports/import_aliases/test.sol", VERSIONS_08),
    # 0.8.9 crashes on our testcase
    Test(
        "user_defined_value_type/user_defined_types-0.8.8.sol", ["0.8.8"] + make_version(8, 10, 15)
    ),
    Test("user_defined_value_type/argument-0.8.8.sol", ["0.8.8"] + make_version(8, 10, 15)),
    Test("user_defined_value_type/calldata-0.8.8.sol", ["0.8.8"] + make_version(8, 10, 15)),
    Test("user_defined_value_type/constant-0.8.8.sol", ["0.8.8"] + make_version(8, 10, 15)),
    Test("user_defined_value_type/erc20-0.8.8.sol", ["0.8.8"] + make_version(8, 10, 15)),
    Test("user_defined_value_type/in_parenthesis-0.8.8.sol", ["0.8.8"] + make_version(8, 10, 15)),
    Test("user_defined_value_type/top-level-0.8.8.sol", ["0.8.8"] + make_version(8, 10, 15)),
    Test("user_defined_value_type/using-for-0.8.8.sol", ["0.8.8"] + make_version(8, 10, 15)),
    Test("user_defined_value_type/abi-decode-fixed-array.sol", ["0.8.8"] + make_version(8, 10, 15)),
    Test("bytes_call.sol", ["0.8.12"]),
    Test("modifier_identifier_path.sol", VERSIONS_08),
    Test("free_functions/libraries_from_free.sol", ["0.8.12"]),
    Test("free_functions/new_operator.sol", ["0.8.12"]),
    Test("free_functions/library_constant_function_collision.sol", ["0.8.12"]),
    Test("ternary-with-max.sol", ["0.8.15"]),
    Test("using-for-1-0.8.0.sol", ["0.8.15"]),
    Test("using-for-2-0.8.0.sol", ["0.8.15"]),
    Test("using-for-3-0.8.0.sol", ["0.8.15"]),
    Test("using-for-4-0.8.0.sol", ["0.8.15"]),
    Test("using-for-in-library-0.8.0.sol", ["0.8.15"]),
    Test("using-for-alias-contract-0.8.0.sol", ["0.8.15"]),
    Test("using-for-alias-top-level-0.8.0.sol", ["0.8.15"]),
    Test("using-for-functions-list-1-0.8.0.sol", ["0.8.15"]),
    Test("using-for-functions-list-2-0.8.0.sol", ["0.8.15"]),
    Test("using-for-functions-list-3-0.8.0.sol", ["0.8.15"]),
    Test("using-for-functions-list-4-0.8.0.sol", ["0.8.15"]),
    Test("using-for-global-0.8.0.sol", ["0.8.15"]),
    Test("using-for-this-contract.sol", ["0.8.15"]),
    Test("library_event-0.8.16.sol", ["0.8.16"]),
    Test("top-level-struct-0.8.0.sol", ["0.8.0"]),
    Test("yul-top-level-0.8.0.sol", ["0.8.0"]),
    Test("complex_imports/import_aliases_issue_1319/test.sol", ["0.5.12"]),
    Test("yul-state-constant-access.sol", ["0.8.16"]),
    Test("negate-unary-element.sol", ["0.8.16"]),
    Test(
        "assembly-functions.sol",
        ["0.6.9", "0.7.6", "0.8.16"],
    ),
    Test("user_defined_operators-0.8.19.sol", ["0.8.19"]),
    Test("aliasing/main.sol", ["0.8.19"]),
    Test("aliasing/alias-unit-NewContract.sol", ["0.8.19"]),
    Test("aliasing/alias-symbol-NewContract.sol", ["0.8.19"]),
    Test("type-aliases.sol", ["0.8.19"]),
    Test("enum-max-min.sol", ["0.8.19"]),
    Test("event-top-level.sol", ["0.8.22"]),
    Test("solidity-0.8.24.sol", ["0.8.24"], solc_args="--evm-version cancun"),
    Test("scope/inherited_function_scope.sol", ["0.8.24"]),
    Test("using_for_global_user_defined_operator_1.sol", ["0.8.24"]),
    Test("require-error.sol", ["0.8.27"]),
    Test("yul-solady.sol", ["0.8.27"]),
]
# create the output folder if needed
try:
    os.mkdir("test_artifacts")
except OSError:
    pass


def pytest_generate_tests(metafunc):
    test_cases = []
    for test_item in ALL_TESTS:
        for version, flavor, _ in test_item.versions_with_flavors:
            test_cases.append((test_item.test_file, version, flavor))
    metafunc.parametrize("test_file, version, flavor", test_cases)


class TestASTParsing:
    def test_parsing(self, test_file, version, flavor):
        actual = os.path.join(TEST_ROOT, "compile", f"{test_file}-{version}-{flavor}.zip")
        expected = os.path.join(TEST_ROOT, "expected", f"{test_file}-{version}-{flavor}.json")

        cc = load_from_zip(actual)[0]

        # Validate that the AST is in the expected format
        for compiled in cc.compilation_units.values():
            for source in compiled.source_units.values():
                if source.ast:
                    if flavor == "compact":
                        assert "nodeType" in source.ast, "AST is not compact"
                    else:
                        assert "nodeType" not in source.ast, "AST is not legacy"

        sl = Slither(
            cc,
            solc_force_legacy_json=flavor == "legacy",
            disallow_partial=True,
            skip_analyze=True,
        )

        actual = generate_output(sl)

        assert os.path.isfile(expected), f"Expected file {expected} does not exist"
        with open(expected, encoding="utf8") as f:
            expected = json.load(f)

        diff = DeepDiff(expected, actual, ignore_order=True, verbose_level=2, view="tree")
        if diff:
            for change in diff.get("values_changed", []):
                path_list = re.findall(r"\['(.*?)'\]", change.path())
                path = "_".join(path_list)
                with open(
                    f"test_artifacts/{test_file}_{path}_expected.dot",
                    "w",
                    encoding="utf8",
                ) as f:
                    f.write(change.t1)
                with open(
                    f"test_artifacts/{test_file}_{version}_{flavor}_{path}_actual.dot",
                    "w",
                    encoding="utf8",
                ) as f:
                    f.write(change.t2)

        assert not diff, diff.pretty()

        sl = Slither(cc, solc_force_legacy_json=flavor == "legacy", disallow_partial=True)
        sl.register_printer(Echidna)
        sl.run_printers()


def _generate_test(test_item: Test, skip_existing=False):
    flavors = ["compact"]
    if not test_item.disable_legacy:
        flavors += ["legacy"]
    for version, flavor, _ in test_item.versions_with_flavors:
        test_file = os.path.join(
            TEST_ROOT, "compile", f"{test_item.test_file}-{version}-{flavor}.zip"
        )
        expected_file = os.path.join(
            TEST_ROOT, "expected", f"{test_item.test_file}-{version}-{flavor}.json"
        )

        if skip_existing:
            if os.path.isfile(expected_file):
                continue

        try:
            cc = load_from_zip(test_file)[0]
            sl = Slither(
                cc,
                solc_force_legacy_json=flavor == "legacy",
                disallow_partial=True,
                skip_analyze=True,
            )

        except Exception as e:
            print(e)
            print(test_item)
            print(f"{expected_file} failed")
            continue

        actual = generate_output(sl)
        print(f"Generate {expected_file}")

        Path(expected_file).parents[0].mkdir(parents=True, exist_ok=True)

        with open(expected_file, "w", encoding="utf8") as f:
            json.dump(actual, f, indent="  ")


def set_solc(version: str):
    env = dict(os.environ)
    env["SOLC_VERSION"] = version
    os.environ.clear()
    os.environ.update(env)


def _generate_compile(test_item: Test, skip_existing=False):
    for version, flavor, solc_args in test_item.versions_with_flavors:
        test_file = os.path.join(TEST_ROOT, test_item.test_file)
        expected_file = os.path.join(
            TEST_ROOT, "compile", f"{test_item.test_file}-{version}-{flavor}.zip"
        )

        if skip_existing:
            if os.path.isfile(expected_file):
                continue

        set_solc(version)
        print(f"Compiled to {expected_file}")
        cc = CryticCompile(
            test_file, solc_force_legacy_json=flavor == "legacy", solc_args=solc_args
        )

        Path(expected_file).parents[0].mkdir(parents=True, exist_ok=True)

        save_to_zip([cc], expected_file)


if __name__ == "__main__":
    required_solcs = set()
    for test in ALL_TESTS:
        required_solcs |= set(test.solc_versions)
    installed_solcs = set(get_installed_solc_versions())
    missing_solcs = list(required_solcs - installed_solcs)
    if missing_solcs:
        install_solc_versions(missing_solcs)

    if len(sys.argv) != 2 or sys.argv[1] not in ["--generate", "--overwrite", "--compile"]:
        print(
            "To generate the missing json artifacts run\n\tpython tests/test_ast_parsing.py --generate"
        )
        print(
            "To re-generate all the json artifacts run\n\tpython tests/test_ast_parsing.py --overwrite"
        )
        print("To compile json artifacts run\n\tpython tests/test_ast_parsing.py --compile")
        print("\tThis will overwrite the previous json files.")
    elif sys.argv[1] == "--generate":
        for next_test in ALL_TESTS:
            _generate_test(next_test, skip_existing=True)
    elif sys.argv[1] == "--overwrite":
        for next_test in ALL_TESTS:
            _generate_test(next_test)
    elif sys.argv[1] == "--compile":
        for next_test in ALL_TESTS:
            _generate_compile(next_test, skip_existing=True)



import logging
from pathlib import Path

from slither.core.declarations import Contract
from slither.tools.properties.addresses.address import Addresses
from slither.tools.properties.platforms.truffle import (
    generate_migration,
    generate_unit_test,
)
from slither.tools.properties.properties.ercs.erc20.properties.initialization import ERC20_CONFIG
from slither.tools.properties.properties.properties import Property

logger = logging.getLogger("Slither")


def generate_truffle_test(
    contract: Contract,
    type_property: str,
    unit_tests: list[Property],
    addresses: Addresses,
) -> str:
    test_contract = f"Test{contract.name}{type_property}"
    filename_init = f"Initialization{test_contract}.js"
    filename = f"{test_contract}.js"

    output_dir = Path(contract.compilation_unit.core.crytic_compile.target)

    generate_migration(test_contract, output_dir, addresses.owner)

    generate_unit_test(
        test_contract,
        filename_init,
        ERC20_CONFIG,
        output_dir,
        addresses,
        f"Check the constructor of {test_contract}",
    )

    generate_unit_test(
        test_contract,
        filename,
        unit_tests,
        output_dir,
        addresses,
    )

    log_info = "\n"
    log_info += "To run the unit tests:\n"
    log_info += f"\ttruffle test {Path(output_dir, 'test', 'crytic', filename_init)}\n"
    log_info += f"\ttruffle test {Path(output_dir, 'test', 'crytic', filename)}\n"
    return log_info



from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.output import Output


class CFG(AbstractPrinter):
    ARGUMENT = "cfg"
    HELP = "Export the CFG of each functions"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#cfg"

    def output(self, filename: str) -> Output:
        """
        _filename is not used
        Args:
            _filename(string)
        """

        info = ""
        all_files = []
        for contract in self.contracts:  # type: ignore
            for function in contract.functions + list(contract.modifiers):
                if filename:
                    new_filename = f"{filename}-{contract.name}-{function.full_name}.dot"
                else:
                    new_filename = f"{contract.name}-{function.full_name}.dot"
                info += f"Export {new_filename}\n"
                content = function.slithir_cfg_to_dot_str()
                with open(new_filename, "w", encoding="utf8") as f:
                    f.write(content)
                all_files.append((new_filename, content))

        self.info(info)

        res = self.generate_output(info)
        for filename_result, content in all_files:
            res.add_file(filename_result, content)
        return res



import re

from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.utils.output import Output


class RightToLeftOverride(AbstractDetector):
    """
    Detect the usage of a Right-To-Left-Override (U+202E) character
    """

    ARGUMENT = "rtlo"
    HELP = "Right-To-Left-Override control character is used"
    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#right-to-left-override-character"
    WIKI_TITLE = "Right-to-Left-Override character"
    WIKI_DESCRIPTION = "An attacker can manipulate the logic of the contract by using a right-to-left-override character (`U+202E)`."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Token
{

    address payable o; // owner
    mapping(address => uint) tokens;

    function withdraw() external returns(uint)
    {
        uint amount = tokens[msg.sender];
        address payable d = msg.sender;
        tokens[msg.sender] = 0;
        _withdraw(/*owner‮/*noitanitsed*/ d, o/*‭
		        /*value */, amount);
    }

    function _withdraw(address payable fee_receiver, address payable destination, uint value) internal
    {
		fee_receiver.transfer(1);
		destination.transfer(value);
    }
}
```

`Token` uses the right-to-left-override character when calling `_withdraw`. As a result, the fee is incorrectly sent to `msg.sender`, and the token balance is sent to the owner.

"""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Special control characters must not be allowed."

    RTLO_CHARACTER_ENCODED = "\u202e".encode()
    STANDARD_JSON = False

    def _detect(self) -> list[Output]:
        results = []
        pattern = re.compile(".*\u202e.*".encode())

        for filename, source in self.slither.source_code.items():
            # Attempt to find all RTLO characters in this source file.
            original_source_encoded = source.encode("utf8")
            start_index = 0

            # Keep searching all file contents for the character.
            while True:
                source_encoded = original_source_encoded[start_index:]
                result_index = source_encoded.find(self.RTLO_CHARACTER_ENCODED)

                # If we couldn't find the character in the remainder of source, stop.
                if result_index == -1:
                    break

                # We found another instance of the character, define our output
                idx = start_index + result_index

                relative = self.slither.crytic_compile.filename_lookup(filename).relative
                info: DETECTOR_INFO = f"{relative} contains a unicode right-to-left-override character at byte offset {idx}:\n"

                # We have a patch, so pattern.find will return at least one result

                info += f"\t- {pattern.findall(source_encoded)[0]}\n"
                res = self.generate_result(info)
                res.add_other(
                    "rtlo-character",
                    (filename, idx, len(self.RTLO_CHARACTER_ENCODED)),
                    self.compilation_unit,
                )
                results.append(res)

                # Advance the start index for the next iteration
                start_index = idx + 1

        return results



from slither.tools.upgradeability.checks.abstract_checks import (
    AbstractCheck,
    CheckClassification,
    CHECK_INFO,
)
from slither.utils.output import Output


class WereConstant(AbstractCheck):
    ARGUMENT = "were-constant"
    IMPACT = CheckClassification.HIGH

    HELP = "Variables that should be constant"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#variables-that-should-be-constant"
    WIKI_TITLE = "Variables that should be constant"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect state variables that should be `constant̀`.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Contract{
    uint variable1;
    uint constant variable2;
    uint variable3;
}

contract ContractV2{
    uint variable1;
    uint variable2;
    uint variable3;
}
```
Because `variable2` is not anymore a `constant`, the storage location of `variable3` will be different.
As a result, `ContractV2` will have a corrupted storage layout.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Do not remove `constant` from a state variables during an update.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True
    REQUIRE_CONTRACT_V2 = True

    def _check(self) -> list[Output]:
        contract_v1 = self.contract
        contract_v2 = self.contract_v2

        if contract_v2 is None:
            raise Exception("were-constant requires a V2 contract")
        state_variables_v1 = contract_v1.state_variables
        state_variables_v2 = contract_v2.state_variables

        v2_additional_variables = len(state_variables_v2) - len(state_variables_v1)
        v2_additional_variables = max(v2_additional_variables, 0)

        # We keep two index, because we need to have them out of sync if v2
        # has additional non constant variables
        idx_v1 = 0
        idx_v2 = 0

        results = []
        while idx_v1 < len(state_variables_v1):
            state_v1 = contract_v1.state_variables[idx_v1]
            if len(state_variables_v2) <= idx_v2:
                break

            state_v2 = contract_v2.state_variables[idx_v2]

            if state_v2:
                if state_v1.is_constant:
                    if not state_v2.is_constant:
                        # If v2 has additional non constant variables, we need to skip them
                        if (
                            state_v1.name != state_v2.name or state_v1.type != state_v2.type
                        ) and v2_additional_variables > 0:
                            v2_additional_variables -= 1
                            idx_v2 += 1
                            continue
                        info: CHECK_INFO = [state_v1, " was constant, but ", state_v2, "is not.\n"]
                        json = self.generate_result(info)
                        results.append(json)

            idx_v1 += 1
            idx_v2 += 1

        return results


class BecameConstant(AbstractCheck):
    ARGUMENT = "became-constant"
    IMPACT = CheckClassification.HIGH

    HELP = "Variables that should not be constant"
    WIKI = "https://github.com/crytic/slither/wiki/Upgradeability-Checks#variables-that-should-not-be-constant"
    WIKI_TITLE = "Variables that should not be constant"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detect state variables that should not be `constant̀`.
"""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract Contract{
    uint variable1;
    uint variable2;
    uint variable3;
}

contract ContractV2{
    uint variable1;
    uint constant variable2;
    uint variable3;
}
```
Because `variable2` is now a `constant`, the storage location of `variable3` will be different.
As a result, `ContractV2` will have a corrupted storage layout.
"""
    # endregion wiki_exploit_scenario

    # region wiki_recommendation
    WIKI_RECOMMENDATION = """
Do not make an existing state variable `constant`.
"""
    # endregion wiki_recommendation

    REQUIRE_CONTRACT = True
    REQUIRE_CONTRACT_V2 = True

    def _check(self) -> list[Output]:
        contract_v1 = self.contract
        contract_v2 = self.contract_v2

        if contract_v2 is None:
            raise Exception("became-constant requires a V2 contract")

        state_variables_v1 = contract_v1.state_variables
        state_variables_v2 = contract_v2.state_variables

        v2_additional_variables = len(state_variables_v2) - len(state_variables_v1)
        v2_additional_variables = max(v2_additional_variables, 0)

        # We keep two index, because we need to have them out of sync if v2
        # has additional non constant variables
        idx_v1 = 0
        idx_v2 = 0

        results = []
        while idx_v1 < len(state_variables_v1):
            state_v1 = contract_v1.state_variables[idx_v1]
            if len(state_variables_v2) <= idx_v2:
                break

            state_v2 = contract_v2.state_variables[idx_v2]

            if state_v2:
                if state_v1.is_constant:
                    if not state_v2.is_constant:
                        # If v2 has additional non constant variables, we need to skip them
                        if (
                            state_v1.name != state_v2.name or state_v1.type != state_v2.type
                        ) and v2_additional_variables > 0:
                            v2_additional_variables -= 1
                            idx_v2 += 1
                            continue
                elif state_v2.is_constant:
                    info: CHECK_INFO = [state_v1, " was not constant but ", state_v2, " is.\n"]
                    json = self.generate_result(info)
                    results.append(json)

            idx_v1 += 1
            idx_v2 += 1

        return results



from slither.core.expressions import Literal
from slither.core.variables.variable import Variable
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator
from slither.tools.mutator.utils.patch import create_patch_with_line


class MVIE(AbstractMutator):
    NAME = "MVIE"
    HELP = "variable initialization using an expression"

    def _mutate(self) -> dict:
        result: dict = {}
        variable: Variable

        # Create fault for state variables declaration
        for variable in self.contract.state_variables_declared:
            if variable.initialized:
                # Cannot remove the initialization of constant variables
                if variable.is_constant:
                    continue

                if not isinstance(variable.expression, Literal):
                    # Get the string
                    start = variable.source_mapping.start
                    stop = variable.expression.source_mapping.start
                    old_str = variable.source_mapping.content
                    new_str = old_str[: old_str.find("=")]
                    line_no = variable.node_initialization.source_mapping.lines
                    if line_no[0] not in self.dont_mutate_line:
                        create_patch_with_line(
                            result,
                            self.in_file,
                            start,
                            stop + variable.expression.source_mapping.length,
                            old_str,
                            new_str,
                            line_no[0],
                        )

        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            for variable in function.local_variables:
                if variable.initialized and not isinstance(variable.expression, Literal):
                    # Get the string
                    start = variable.source_mapping.start
                    stop = variable.expression.source_mapping.start
                    old_str = variable.source_mapping.content
                    new_str = old_str[: old_str.find("=")]
                    line_no = variable.source_mapping.lines
                    if line_no[0] not in self.dont_mutate_line:
                        create_patch_with_line(
                            result,
                            self.in_file,
                            start,
                            stop + variable.expression.source_mapping.length,
                            old_str,
                            new_str,
                            line_no[0],
                        )
        return result



"""
Halstead complexity metrics
https://en.wikipedia.org/wiki/Halstead_complexity_measures

12 metrics based on the number of unique operators and operands:

Core metrics:
n1 = the number of distinct operators
n2 = the number of distinct operands
N1 = the total number of operators
N2 = the total number of operands

Extended metrics1:
n = n1 + n2  # Program vocabulary
N = N1 + N2  # Program length
S = n1 * log2(n1) + n2 * log2(n2) # Estimated program length
V = N * log2(n) # Volume

Extended metrics2:
D = (n1 / 2) * (N2 / n2) # Difficulty
E = D * V # Effort
T = E / 18 seconds # Time required to program
B = (E^(2/3)) / 3000 # Number of delivered bugs

"""

from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.halstead import HalsteadMetrics
from slither.utils.output import Output


class Halstead(AbstractPrinter):
    ARGUMENT = "halstead"
    HELP = "Computes the Halstead complexity metrics for each contract"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#halstead"

    def output(self, _filename: str) -> Output:
        if len(self.contracts) == 0:
            return self.generate_output("No contract found")

        halstead = HalsteadMetrics(self.contracts)

        res = self.generate_output(halstead.full_text)
        res.add_pretty_table(halstead.core.pretty_table, halstead.core.title)
        res.add_pretty_table(halstead.extended1.pretty_table, halstead.extended1.title)
        res.add_pretty_table(halstead.extended2.pretty_table, halstead.extended2.title)
        self.info(halstead.full_text)

        return res



import json
from collections import defaultdict
from typing import NamedTuple

from slither.analyses.data_dependency.data_dependency import is_dependent
from slither.core.cfg.node import Node
from slither.core.declarations import Enum, Function, Contract
from slither.core.declarations.solidity_variables import (
    SolidityVariableComposed,
    SolidityFunction,
    SolidityVariable,
)
from slither.core.expressions import NewContract
from slither.core.slither_core import SlitherCore
from slither.core.solidity_types import TypeAlias
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.core.variables.state_variable import StateVariable
from slither.core.variables.variable import Variable
from slither.printers.abstract_printer import AbstractPrinter
from slither.slithir.operations import (
    Member,
    Operation,
    SolidityCall,
    LowLevelCall,
    HighLevelCall,
    EventCall,
    Send,
    Transfer,
    InternalDynamicCall,
    InternalCall,
    TypeConversion,
)
from slither.slithir.operations.binary import Binary
from slither.slithir.variables import Constant, ReferenceVariable
from slither.utils.output import Output
from slither.visitors.expression.constants_folding import ConstantFolding, NotConstant


def _get_name(f: Function | Variable) -> str:
    # Return the name of the function or variable
    if isinstance(f, Function):
        if f.is_fallback or f.is_receive:
            return "()"
    return f.solidity_signature


def _extract_payable(contracts: list[Contract]) -> dict[str, list[str]]:
    ret: dict[str, list[str]] = {}
    for contract in contracts:
        payable_functions = [_get_name(f) for f in contract.functions_entry_points if f.payable]
        if payable_functions:
            ret[contract.name] = payable_functions
    return ret


def _extract_solidity_variable_usage(
    contracts: list[Contract], sol_var: SolidityVariable
) -> dict[str, list[str]]:
    ret: dict[str, list[str]] = {}
    for contract in contracts:
        functions_using_sol_var = []
        for f in contract.functions_entry_points:
            for v in f.all_solidity_variables_read():
                if v == sol_var:
                    functions_using_sol_var.append(_get_name(f))
                    break
        if functions_using_sol_var:
            ret[contract.name] = functions_using_sol_var
    return ret


def _is_constant(f: Function) -> bool:
    """
    Heuristic:
    - If view/pure with Solidity >= 0.4 -> Return true
    - If it contains assembly -> Return false (SlitherCore doesn't analyze asm)
    - Otherwise check for the rules from
    https://solidity.readthedocs.io/en/v0.5.0/contracts.html?highlight=pure#view-functions
    with an exception: internal dynamic call are not correctly handled, so we consider them as non-constant
    :param f:
    :return:
    """
    if f.view or f.pure:
        if not f.compilation_unit.solc_version.startswith("0.4"):
            return True
    if f.payable:
        return False
    if not f.is_implemented:
        return False
    if f.contains_assembly:
        return False
    if f.all_state_variables_written():
        return False
    for ir in f.all_slithir_operations():
        if isinstance(ir, InternalDynamicCall):
            return False
        if isinstance(ir, (EventCall, NewContract, LowLevelCall, Send, Transfer)):
            return False
        if isinstance(ir, SolidityCall) and ir.function in [
            SolidityFunction("selfdestruct(address)"),
            SolidityFunction("suicide(address)"),
        ]:
            return False
        if isinstance(ir, HighLevelCall):
            if isinstance(ir.function, Variable) or ir.function.view or ir.function.pure:
                # External call to constant functions are ensured to be constant only for solidity >= 0.5
                if f.compilation_unit.solc_version.startswith("0.4"):
                    return False
            else:
                return False
        if isinstance(ir, InternalCall) and ir.function:
            # Storage write are not properly handled by all_state_variables_written
            if any(parameter.is_storage for parameter in ir.function.parameters):
                return False
    return True


def _extract_constant_functions(contracts: list[Contract]) -> dict[str, list[str]]:
    ret: dict[str, list[str]] = {}
    for contract in contracts:
        cst_functions = [_get_name(f) for f in contract.functions_entry_points if _is_constant(f)]
        cst_functions += [
            v.solidity_signature for v in contract.state_variables if v.visibility in ["public"]
        ]
        if cst_functions:
            ret[contract.name] = cst_functions
    return ret


def _extract_assert(contracts: list[Contract]) -> dict[str, dict[str, list[dict]]]:
    """
    Return the list of contract -> function name -> List(source mapping of the assert))

    Args:
        contracts: list of contracts

    Returns:

    """
    ret: dict[str, dict[str, list[dict]]] = {}
    for contract in contracts:
        functions_using_assert: dict[str, list[dict]] = defaultdict(list)
        for f in contract.functions_entry_points:
            for ir in f.all_solidity_calls():
                if ir.function == SolidityFunction("assert(bool)") and ir.node.source_mapping:
                    func_name = _get_name(f)
                    functions_using_assert[func_name].append(ir.node.source_mapping.to_json())
        if functions_using_assert:
            ret[contract.name] = functions_using_assert
    return ret


# Create a named tuple that is serialization in json
def json_serializable(cls):
    my_super = super

    def as_dict(self):
        yield dict(zip(self._fields, my_super(cls, self).__iter__(), strict=False))

    cls.__iter__ = as_dict
    return cls


@json_serializable
class ConstantValue(NamedTuple):
    # Here value should be  Union[str, int, bool]
    # But the json lib in Echidna does not handle large integer in json
    # So we convert everything to string
    value: str
    type: str


def _extract_constant_from_read(
    ir: Operation,
    r: SourceMapping,
    all_cst_used: list[ConstantValue],
    all_cst_used_in_binary: dict[str, list[ConstantValue]],
    context_explored: set[Node],
) -> None:
    var_read = r.points_to_origin if isinstance(r, ReferenceVariable) else r
    # Do not report struct_name in a.struct_name
    if isinstance(ir, Member):
        return
    if isinstance(var_read, Variable) and var_read.is_constant:
        # In case of type conversion we use the destination type
        if isinstance(ir, TypeConversion):
            if isinstance(ir.type, TypeAlias):
                value_type = ir.type.type
            else:
                value_type = ir.type
        else:
            value_type = var_read.type
        try:
            value = ConstantFolding(var_read.expression, value_type).result()
            all_cst_used.append(ConstantValue(str(value), str(value_type)))
        except NotConstant:
            pass
    if isinstance(var_read, Constant):
        all_cst_used.append(ConstantValue(str(var_read.value), str(var_read.type)))
    if isinstance(var_read, StateVariable):
        if var_read.node_initialization:
            if var_read.node_initialization.irs:
                if var_read.node_initialization in context_explored:
                    return
                context_explored.add(var_read.node_initialization)
                _extract_constants_from_irs(
                    var_read.node_initialization.irs,
                    all_cst_used,
                    all_cst_used_in_binary,
                    context_explored,
                )


def _extract_constant_from_binary(
    ir: Binary,
    all_cst_used: list[ConstantValue],
    all_cst_used_in_binary: dict[str, list[ConstantValue]],
):
    for r in ir.read:
        if isinstance(r, Constant):
            all_cst_used_in_binary[str(ir.type)].append(ConstantValue(str(r.value), str(r.type)))
        if isinstance(ir.variable_left, Constant) or isinstance(ir.variable_right, Constant):
            if ir.lvalue:
                try:
                    type_ = ir.lvalue.type
                    cst = ConstantFolding(ir.expression, type_).result()
                    all_cst_used.append(ConstantValue(str(cst.value), str(type_)))
                except NotConstant:
                    pass


def _extract_constants_from_irs(
    irs: list[Operation],
    all_cst_used: list[ConstantValue],
    all_cst_used_in_binary: dict[str, list[ConstantValue]],
    context_explored: set[Node],
) -> None:
    for ir in irs:
        if isinstance(ir, Binary):
            _extract_constant_from_binary(ir, all_cst_used, all_cst_used_in_binary)
        if isinstance(ir, TypeConversion):
            if isinstance(ir.variable, Constant):
                if isinstance(ir.type, TypeAlias):
                    value_type = ir.type.type
                else:
                    value_type = ir.type
                all_cst_used.append(ConstantValue(str(ir.variable.value), str(value_type)))
                continue
        if (
            isinstance(ir, Member)
            and isinstance(ir.variable_left, Enum)
            and isinstance(ir.variable_right, Constant)
        ):
            # enums are constant values
            try:
                internal_num = ir.variable_left.values.index(ir.variable_right.value)
                all_cst_used.append(ConstantValue(str(internal_num), "uint256"))
            except ValueError:  # index could fail; should never happen in working solidity code
                pass
        for r in ir.read:
            _extract_constant_from_read(
                ir, r, all_cst_used, all_cst_used_in_binary, context_explored
            )


def _extract_constants(
    contracts: list[Contract],
) -> tuple[dict[str, dict[str, list]], dict[str, dict[str, dict]]]:
    # contract -> function -> [ {"value": value, "type": type} ]
    ret_cst_used: dict[str, dict[str, list[ConstantValue]]] = defaultdict(dict)
    # contract -> function -> binary_operand -> [ {"value": value, "type": type ]
    ret_cst_used_in_binary: dict[str, dict[str, dict[str, list[ConstantValue]]]] = defaultdict(dict)
    for contract in contracts:
        for function in contract.functions_entry_points:
            all_cst_used: list = []
            all_cst_used_in_binary: dict = defaultdict(list)

            context_explored = set()
            context_explored.add(function)
            _extract_constants_from_irs(
                function.all_slithir_operations(),
                all_cst_used,
                all_cst_used_in_binary,
                context_explored,
            )

            # Note: use list(set()) instead of set
            # As this is meant to be serialized in JSON, and JSON does not support set
            if all_cst_used:
                ret_cst_used[contract.name][_get_name(function)] = list(set(all_cst_used))
            if all_cst_used_in_binary:
                ret_cst_used_in_binary[contract.name][_get_name(function)] = {
                    k: list(set(v)) for k, v in all_cst_used_in_binary.items()
                }
    return ret_cst_used, ret_cst_used_in_binary


def _extract_function_relations(
    contracts: list[Contract],
) -> dict[str, dict[str, dict[str, list[str]]]]:
    # contract -> function -> [functions]
    ret: dict[str, dict[str, dict[str, list[str]]]] = defaultdict(dict)
    for contract in contracts:
        ret[contract.name] = defaultdict(dict)
        written = {
            _get_name(function): function.all_state_variables_written()
            for function in contract.functions_entry_points
        }
        read = {
            _get_name(function): function.all_state_variables_read()
            for function in contract.functions_entry_points
        }
        for function in contract.functions_entry_points:
            ret[contract.name][_get_name(function)] = {
                "impacts": [],
                "is_impacted_by": [],
            }
            for candidate, varsWritten in written.items():
                if any(r in varsWritten for r in function.all_state_variables_read()):
                    ret[contract.name][_get_name(function)]["is_impacted_by"].append(candidate)
            for candidate, varsRead in read.items():
                if any(r in varsRead for r in function.all_state_variables_written()):
                    ret[contract.name][_get_name(function)]["impacts"].append(candidate)
    return ret


def _have_external_calls(contracts: list[Contract]) -> dict[str, list[str]]:
    """
    Detect the functions with external calls
    :param slither:
    :return:
    """
    ret: dict[str, list[str]] = defaultdict(list)
    for contract in contracts:
        for function in contract.functions_entry_points:
            if function.all_high_level_calls() or function.all_low_level_calls():
                ret[contract.name].append(_get_name(function))
        if contract.name in ret:
            ret[contract.name] = list(set(ret[contract.name]))
    return ret


def _use_balance(contracts: list[Contract]) -> dict[str, list[str]]:
    """
    Detect the functions with external calls
    :param slither:
    :return:
    """
    ret: dict[str, list[str]] = defaultdict(list)
    for contract in contracts:
        for function in contract.functions_entry_points:
            for ir in function.all_slithir_operations():
                if isinstance(ir, SolidityCall) and ir.function == SolidityFunction(
                    "balance(address)"
                ):
                    ret[contract.name].append(_get_name(function))
        if contract.name in ret:
            ret[contract.name] = list(set(ret[contract.name]))
    return ret


def _with_fallback(contracts: list[Contract]) -> set[str]:
    ret: set[str] = set()
    for contract in contracts:
        for function in contract.functions_entry_points:
            if function.is_fallback:
                ret.add(contract.name)
    return ret


def _with_receive(contracts: list[Contract]) -> set[str]:
    ret: set[str] = set()
    for contract in contracts:
        for function in contract.functions_entry_points:
            if function.is_receive:
                ret.add(contract.name)
    return ret


def _call_a_parameter(slither: SlitherCore, contracts: list[Contract]) -> dict[str, list[dict]]:
    """
    Detect the functions with external calls
    :param slither:
    :return:
    """
    # contract -> [ (function, idx, interface_called) ]
    ret: dict[str, list[dict]] = defaultdict(list)
    for contract in contracts:
        for function in contract.functions_entry_points:
            try:
                for ir in function.all_slithir_operations():
                    if isinstance(ir, HighLevelCall):
                        for idx, parameter in enumerate(function.parameters):
                            if is_dependent(ir.destination, parameter, function):
                                ret[contract.name].append(
                                    {
                                        "function": _get_name(function),
                                        "parameter_idx": idx,
                                        "signature": _get_name(ir.function),
                                    }
                                )
                    if isinstance(ir, LowLevelCall):
                        for idx, parameter in enumerate(function.parameters):
                            if is_dependent(ir.destination, parameter, function):
                                ret[contract.name].append(
                                    {
                                        "function": _get_name(function),
                                        "parameter_idx": idx,
                                        "signature": None,
                                    }
                                )
            except Exception as e:
                if slither.no_fail:
                    continue
                raise e
    return ret


class Echidna(AbstractPrinter):
    ARGUMENT = "echidna"
    HELP = "Export Echidna guiding information"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#echidna"

    def output(self, filename: str) -> Output:
        """
        Output the inheritance relation

        _filename is not used
        Args:
            _filename(string)
        """

        contracts = self.slither.contracts

        payable = _extract_payable(contracts)
        timestamp = _extract_solidity_variable_usage(
            contracts, SolidityVariableComposed("block.timestamp")
        )
        block_number = _extract_solidity_variable_usage(
            contracts, SolidityVariableComposed("block.number")
        )
        msg_sender = _extract_solidity_variable_usage(
            contracts, SolidityVariableComposed("msg.sender")
        )
        msg_gas = _extract_solidity_variable_usage(contracts, SolidityVariableComposed("msg.gas"))
        assert_usage = _extract_assert(contracts)
        cst_functions = _extract_constant_functions(contracts)
        (cst_used, cst_used_in_binary) = _extract_constants(contracts)

        functions_relations = _extract_function_relations(contracts)

        constructors = {
            contract.name: contract.constructor.full_name
            for contract in contracts
            if contract.constructor
        }

        external_calls = _have_external_calls(contracts)

        # call_parameters = _call_a_parameter(self.slither, contracts)

        use_balance = _use_balance(contracts)

        with_fallback = list(_with_fallback(contracts))

        with_receive = list(_with_receive(contracts))

        d = {
            "payable": payable,
            "timestamp": timestamp,
            "block_number": block_number,
            "msg_sender": msg_sender,
            "msg_gas": msg_gas,
            "assert": assert_usage,
            "constant_functions": cst_functions,
            "constants_used": cst_used,
            "constants_used_in_binary": cst_used_in_binary,
            "functions_relations": functions_relations,
            "constructors": constructors,
            "have_external_calls": external_calls,
            # "call_a_parameter": call_parameters,
            "use_balance": use_balance,
            "solc_versions": [unit.solc_version for unit in self.slither.compilation_units],
            "with_fallback": with_fallback,
            "with_receive": with_receive,
        }

        self.info(json.dumps(d, indent=4))

        res = self.generate_output(json.dumps(d, indent=4))

        return res



from typing import Union, TYPE_CHECKING, Any

from slither.core.solidity_types import ElementaryType
from slither.core.solidity_types.type import Type

if TYPE_CHECKING:
    from slither.core.declarations.contract import Contract
    from slither.core.declarations.enum import Enum


# Use to model the Type(X) function, which returns an undefined type
# https://solidity.readthedocs.io/en/latest/units-and-global-variables.html#type-information
class TypeInformation(Type):
    def __init__(self, c: Union[ElementaryType, "Contract", "Enum"]) -> None:
        from slither.core.declarations.contract import Contract
        from slither.core.declarations.enum import Enum

        assert isinstance(c, (Contract, ElementaryType, Enum))
        super().__init__()
        self._type = c

    @property
    def type(self) -> Union["Contract", ElementaryType, "Enum"]:
        return self._type

    @property
    def storage_size(self) -> tuple[int, bool]:
        """
        32 is incorrect, as Type(x) return a kind of structure that can contain
        an arbitrary number of value
        As Type(x) cannot be directly stored, we are assuming that the correct storage size
        will be handled by the fields access

        :return:
        """
        return 32, True

    @property
    def is_dynamic(self) -> bool:
        raise NotImplementedError

    def __str__(self) -> str:
        return f"type({self.type.name})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TypeInformation):
            return False
        return self.type == other.type



from slither.detectors.abstract_detector import AbstractDetector, DetectorClassification, Output
from slither.core.scope.scope import FileScope


class UnusedImport(AbstractDetector):
    """
    Detector unused imports.
    """

    ARGUMENT = "unused-import"
    HELP = "Detects unused imports"
    IMPACT = DetectorClassification.INFORMATIONAL
    CONFIDENCE = DetectorClassification.HIGH

    WIKI = "https://github.com/crytic/slither/wiki/Detector-Documentation#unused-imports"

    WIKI_TITLE = "Unused Imports"
    WIKI_DESCRIPTION = "Importing a file that is not used in the contract likely indicates a mistake. The import should be removed until it is needed."

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
    ```solidity
    import {A} from "./A.sol";
    contract B {}
    ```
    B either should import from A and it was forgotten or the import is not needed and should be removed.
    """
    # endregion wiki_exploit_scenario
    WIKI_RECOMMENDATION = (
        "Remove the unused import. If the import is needed later, it can be added back."
    )

    @staticmethod
    def _is_import_container(scope: FileScope) -> bool:
        """
        Returns True if a given file (provided as a `FileScope` object) contains only `import` directives (and pragmas).
        Such a file doesn't need the imports it contains, but its purpose is to aggregate certain correlated imports.
        """
        for c in scope.contracts.values():
            if c.file_scope == scope:
                return False
        for err in scope.custom_errors:
            if err.file_scope == scope:
                return False
        for en in scope.enums.values():
            if en.file_scope == scope:
                return False
        for f in scope.functions:
            if f.file_scope == scope:
                return False
        for st in scope.structures.values():
            if st.file_scope == scope:
                return False
        for ct in scope.type_aliases.values():
            if ct.source_mapping and ct.source_mapping.filename == scope.filename:
                return False
        for uf in scope.using_for_directives:
            if uf.file_scope == scope:
                return False
        for v in scope.variables.values():
            if v.file_scope == scope:
                return False
        return True

    def _detect(self) -> list[Output]:
        results: list[Output] = []
        # This is computed lazily and then memoized so we need to trigger the computation.
        self.slither._compute_offsets_to_ref_impl_decl()

        for unit in self.slither.compilation_units:
            for filename, current_scope in unit.scopes.items():
                # Skip files that are dependencies
                if unit.crytic_compile.is_dependency(filename.absolute):
                    continue

                unused_list = []
                for i in current_scope.imports:
                    # `scope.imports` contains all transitive imports so we need to filter out imports not explicitly imported in the file.
                    # Otherwise, we would recommend removing an import that is used by a leaf contract and cause compilation errors.
                    if i.scope != current_scope:
                        continue

                    # If a scope doesn't define any contract, function, etc., it is an import container.
                    # The second case accounts for importing from an import container as a reference will only be in the definition's file.
                    if self._is_import_container(i.scope) or self._is_import_container(
                        unit.get_scope(i.filename)
                    ):
                        continue

                    imported_path = self.slither.crytic_compile.filename_lookup(i.filename)

                    use_found = False
                    # Search through all references to the imported file
                    for _, refs_to_imported_path in self.slither._offset_to_references[
                        imported_path
                    ].items():
                        for ref in refs_to_imported_path:
                            # If there is a reference in this file to the imported file, it is used.
                            if ref.filename == filename:
                                use_found = True
                                break

                        if use_found:
                            break

                    if not use_found:
                        unused_list.append(f"{i.source_mapping.content} ({i.source_mapping})")

                if len(unused_list) > 0:
                    info = [
                        f"The following unused import(s) in {filename.used} should be removed:",
                    ]
                    for unused in unused_list:
                        info += ["\n\t-", unused, "\n"]

                    results.append(self.generate_result(info))

        return results



from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from slither.core.cfg.node import Node
    from slither.core.declarations.function import Function


class Scope:
    def __init__(
        self, is_checked: bool, is_yul: bool, parent_scope: Union["Scope", "Function"]
    ) -> None:
        self.nodes: list[Node] = []
        self.is_checked = is_checked
        self.is_yul = is_yul
        self.father = parent_scope



from slither.analyses.data_dependency.data_dependency import get_dependencies
from slither.core.cfg.node import Node, NodeType
from slither.core.declarations import (
    Contract,
    Function,
)
from slither.core.expressions import (
    Literal,
    Identifier,
    CallExpression,
    AssignmentOperation,
)
from slither.core.solidity_types import (
    ElementaryType,
)
from slither.core.variables.local_variable import LocalVariable
from slither.core.variables.state_variable import StateVariable
from slither.core.variables.variable import Variable
from slither.slithir.operations import (
    LowLevelCall,
)
from slither.tools.read_storage.read_storage import SlotInfo, SlitherReadStorage
from slither.utils.encoding import encode_ir_for_upgradeability_compare


class TaintedExternalContract:
    def __init__(self, contract: "Contract") -> None:
        self._contract: Contract = contract
        self._tainted_functions: list[Function] = []
        self._tainted_variables: list[Variable] = []

    @property
    def contract(self) -> Contract:
        return self._contract

    @property
    def tainted_functions(self) -> list[Function]:
        return self._tainted_functions

    def add_tainted_function(self, f: Function):
        self._tainted_functions.append(f)

    @property
    def tainted_variables(self) -> list[Variable]:
        return self._tainted_variables

    def add_tainted_variable(self, v: Variable):
        self._tainted_variables.append(v)


def compare(
    v1: Contract, v2: Contract, include_external: bool = False
) -> tuple[
    list[Variable],
    list[Variable],
    list[Variable],
    list[Function],
    list[Function],
    list[Function],
]:
    """
    Compares two versions of a contract. Most useful for upgradeable (logic) contracts,
    but does not require that Contract.is_upgradeable returns true for either contract.

    Args:
        v1: Original version of (upgradeable) contract
        v2: Updated version of (upgradeable) contract
        include_external: Optional flag to enable cross-contract external taint analysis

    Returns:
        missing-vars-in-v2: list[Variable],
        new-variables: list[Variable],
        tainted-variables: list[Variable],
        new-functions: list[Function],
        modified-functions: list[Function],
        tainted-functions: list[Function]
        tainted-contracts: list[TaintedExternalContract]
    """

    order_vars1 = v1.storage_variables_ordered + v1.transient_variables_ordered
    order_vars2 = v2.storage_variables_ordered + v2.transient_variables_ordered
    func_sigs1 = [function.solidity_signature for function in v1.functions]
    func_sigs2 = [function.solidity_signature for function in v2.functions]

    missing_vars_in_v2 = []
    new_variables = []
    tainted_variables = []
    new_functions = []
    modified_functions = []
    tainted_functions = []

    # Since this is not a detector, include any missing variables in the v2 contract
    if len(order_vars2) < len(order_vars1):
        missing_vars_in_v2.extend(get_missing_vars(v1, v2))

    # Find all new and modified functions in the v2 contract
    new_modified_functions = []
    new_modified_function_vars = []
    for sig in func_sigs2:
        function = v2.get_function_from_signature(sig)
        orig_function = v1.get_function_from_signature(sig)
        if sig not in func_sigs1:
            new_modified_functions.append(function)
            new_functions.append(function)
            new_modified_function_vars += function.all_state_variables_written()
        elif not function.is_constructor_variables and is_function_modified(
            orig_function, function
        ):
            new_modified_functions.append(function)
            modified_functions.append(function)
            new_modified_function_vars += function.all_state_variables_written()

    # Find all unmodified functions that call a modified function or read/write the
    # same state variable(s) as a new/modified function, i.e., tainted functions
    for function in v2.functions:
        if (
            function in new_modified_functions
            or function.is_constructor
            or function.name.startswith("slither")
        ):
            continue
        modified_calls = [
            func
            for func in new_modified_functions
            if func in [ir.function for ir in function.internal_calls]
        ]
        tainted_vars = [
            var
            for var in set(new_modified_function_vars)
            if var in function.all_state_variables_read() + function.all_state_variables_written()
            and not var.is_constant
            and not var.is_immutable
        ]
        if len(modified_calls) > 0 or len(tainted_vars) > 0:
            tainted_functions.append(function)

    # Find all new or tainted variables, i.e., variables that are written by a new/modified/tainted function
    for var in order_vars2:
        written_by = v2.get_functions_writing_to_variable(var)
        if next((v for v in v1.state_variables_ordered if v.name == var.name), None) is None:
            new_variables.append(var)
        elif any(func in written_by for func in new_modified_functions + tainted_functions):
            tainted_variables.append(var)

    tainted_contracts = []
    if include_external:
        # Find all external contracts and functions called by new/modified/tainted functions
        tainted_contracts = tainted_external_contracts(
            new_functions + modified_functions + tainted_functions
        )

    return (
        missing_vars_in_v2,
        new_variables,
        tainted_variables,
        new_functions,
        modified_functions,
        tainted_functions,
        tainted_contracts,
    )


def tainted_external_contracts(funcs: list[Function]) -> list[TaintedExternalContract]:
    """
    Takes a list of functions from one contract, finds any calls in these to functions in external contracts,
    and determines which variables and functions in the external contracts are tainted by these external calls.
    Args:
        funcs: a list of Function objects to search for external calls.

    Returns:
        TaintedExternalContract() (
            contract: Contract,
            tainted_functions: List[TaintedFunction],
            tainted_variables: List[TaintedVariable]
        )
    """
    tainted_contracts: dict[str, TaintedExternalContract] = {}
    tainted_list: list[TaintedExternalContract] = []

    for func in funcs:
        for contract, ir in func.all_high_level_calls():
            target = ir.function
            if contract.is_library:
                # Not interested in library calls
                continue
            if contract.name not in tainted_contracts:
                # A contract may be tainted by multiple function calls - only make one TaintedExternalContract object
                tainted_contracts[contract.name] = TaintedExternalContract(contract)
            if (
                isinstance(target, Function)
                and target not in funcs
                and target not in (f for f in tainted_contracts[contract.name].tainted_functions)
                and not (target.is_constructor or target.is_fallback or target.is_receive)
            ):
                # Found a high-level call to a new tainted function
                tainted_contracts[contract.name].add_tainted_function(target)
                for var in target.all_state_variables_written():
                    # Consider as tainted all variables written by the tainted function
                    if var not in (v for v in tainted_contracts[contract.name].tainted_variables):
                        tainted_contracts[contract.name].add_tainted_variable(var)
            elif (
                isinstance(target, StateVariable)
                and target not in (v for v in tainted_contracts[contract.name].tainted_variables)
                and target.is_stored
            ):
                # Found a new high-level call to a public state variable getter
                tainted_contracts[contract.name].add_tainted_variable(target)
    for c in tainted_contracts.values():
        tainted_list.append(c)
        contract = c.contract
        variables = c.tainted_variables
        for var in variables:
            # For each tainted variable, consider as tainted any function that reads or writes to it
            read_write = set(
                contract.get_functions_reading_from_variable(var)
                + contract.get_functions_writing_to_variable(var)
            )
            for f in read_write:
                if f not in tainted_contracts[contract.name].tainted_functions and not (
                    f.is_constructor or f.is_fallback or f.is_receive
                ):
                    c.add_tainted_function(f)
    return tainted_list


def tainted_inheriting_contracts(
    tainted_contracts: list[TaintedExternalContract], contracts: list[Contract] | None = None
) -> list[TaintedExternalContract]:
    """
    Takes a list of TaintedExternalContract obtained from tainted_external_contracts, and finds any contracts which
    inherit a tainted contract, as well as any functions that call tainted functions or read tainted variables in
    the inherited contract.
    Args:
        tainted_contracts: the list obtained from `tainted_external_contracts` or `compare`.
        contracts: (optional) the list of contracts to check for inheritance. If not provided, defaults to
                    `contract.compilation_unit.contracts` for each contract in tainted_contracts.

    Returns:
        An updated list of TaintedExternalContract, including all from the input list.
    """
    for tainted in tainted_contracts:
        contract = tainted.contract
        check_contracts = contracts
        if contracts is None:
            check_contracts = contract.compilation_unit.contracts
        # We are only interested in checking contracts that inherit a tainted contract
        check_contracts = [
            c
            for c in check_contracts
            if c.name not in [t.contract.name for t in tainted_contracts]
            and contract.name in [i.name for i in c.inheritance]
        ]
        for c in check_contracts:
            new_taint = TaintedExternalContract(c)
            for f in c.functions_declared:
                # Search for functions that call an inherited tainted function or access an inherited tainted variable
                internal_calls = [
                    ir.function
                    for ir in f.all_internal_calls()
                    if isinstance(ir.function, Function)
                ]
                if any(
                    call.canonical_name == t.canonical_name
                    for t in tainted.tainted_functions
                    for call in internal_calls
                ) or any(
                    var.canonical_name == t.canonical_name
                    for t in tainted.tainted_variables
                    for var in f.all_state_variables_read() + f.all_state_variables_written()
                ):
                    new_taint.add_tainted_function(f)
            for f in new_taint.tainted_functions:
                # For each newly found tainted function, consider as tainted any variable it writes to
                for var in f.all_state_variables_written():
                    if var not in (
                        v for v in tainted.tainted_variables + new_taint.tainted_variables
                    ):
                        new_taint.add_tainted_variable(var)
            for var in new_taint.tainted_variables:
                # For each newly found tainted variable, consider as tainted any function that reads or writes to it
                read_write = set(
                    contract.get_functions_reading_from_variable(var)
                    + contract.get_functions_writing_to_variable(var)
                )
                for f in read_write:
                    if f not in (
                        t for t in tainted.tainted_functions + new_taint.tainted_functions
                    ) and not (f.is_constructor or f.is_fallback or f.is_receive):
                        new_taint.add_tainted_function(f)
            if len(new_taint.tainted_functions) > 0:
                tainted_contracts.append(new_taint)
    return tainted_contracts


def get_missing_vars(v1: Contract, v2: Contract) -> list[StateVariable]:
    """
    Gets all non-constant/immutable StateVariables that appear in v1 but not v2
    Args:
        v1: Contract version 1
        v2: Contract version 2

    Returns:
        List of StateVariables from v1 missing in v2
    """
    results = []
    order_vars1 = v1.storage_variables_ordered + v1.transient_variables_ordered
    order_vars2 = v2.storage_variables_ordered + v2.transient_variables_ordered
    if len(order_vars2) < len(order_vars1):
        for variable in order_vars1:
            if variable.name not in [v.name for v in order_vars2]:
                results.append(variable)
    return results


def is_function_modified(f1: Function, f2: Function) -> bool:
    """
    Compares two versions of a function, and returns True if the function has been modified.
    First checks whether the functions' content hashes are equal to quickly rule out identical functions.
    Walks the CFGs and compares IR operations if hashes differ to rule out false positives, i.e., from changed comments.

    Args:
        f1: Original version of the function
        f2: New version of the function

    Returns:
        True if the functions differ, otherwise False
    """
    # If the function content hashes are the same, no need to investigate the function further
    if f1.source_mapping.content_hash == f2.source_mapping.content_hash:
        return False
    # If the hashes differ, it is possible a change in a name or in a comment could be the only difference
    # So we need to resort to walking through the CFG and comparing the IR operations
    queue_f1 = [f1.entry_point]
    queue_f2 = [f2.entry_point]
    visited = []
    while len(queue_f1) > 0 and len(queue_f2) > 0:
        node_f1 = queue_f1.pop(0)
        node_f2 = queue_f2.pop(0)
        visited.extend([node_f1, node_f2])
        queue_f1.extend(son for son in node_f1.sons if son not in visited)
        queue_f2.extend(son for son in node_f2.sons if son not in visited)
        if len(node_f1.irs) != len(node_f2.irs):
            return True
        for i, ir in enumerate(node_f1.irs):
            if encode_ir_for_upgradeability_compare(ir) != encode_ir_for_upgradeability_compare(
                node_f2.irs[i]
            ):
                return True
    return False


def get_proxy_implementation_slot(proxy: Contract) -> SlotInfo | None:
    """
    Gets information about the storage slot where a proxy's implementation address is stored.
    Args:
        proxy: A Contract object (proxy.is_upgradeable_proxy should be true).

    Returns:
        (`SlotInfo`) | None : A dictionary of the slot information.
    """

    delegate = get_proxy_implementation_var(proxy)
    if isinstance(delegate, StateVariable):
        if delegate.is_stored:
            srs = SlitherReadStorage([proxy], 20)
            return srs.get_storage_slot(delegate, proxy)
        if delegate.is_constant and delegate.type.name == "bytes32":
            return SlotInfo(
                name=delegate.name,
                type_string="address",
                slot=int(delegate.expression.value, 16),
                size=160,
                offset=0,
            )
    return None


def get_proxy_implementation_var(proxy: Contract) -> Variable | None:
    """
    Gets the Variable that stores a proxy's implementation address. Uses data dependency to trace any LocalVariable
    that is passed into a delegatecall as the target address back to its data source, ideally a StateVariable.
    Can return a newly created StateVariable if an `sload` from a hardcoded storage slot is found in assembly.
    Args:
        proxy: A Contract object (proxy.is_upgradeable_proxy should be true).

    Returns:
        (`Variable`) | None : The variable, ideally a StateVariable, which stores the proxy's implementation address.
    """
    if not proxy.is_upgradeable_proxy or not proxy.fallback_function:
        return None

    delegate = find_delegate_in_fallback(proxy)
    if isinstance(delegate, LocalVariable):
        dependencies = get_dependencies(delegate, proxy)
        try:
            delegate = next(var for var in dependencies if isinstance(var, StateVariable))
        except StopIteration:
            # TODO: Handle case where get_dependencies does not return any state variables.
            return delegate
    return delegate


def find_delegate_in_fallback(proxy: Contract) -> Variable | None:
    """
    Searches a proxy's fallback function for a delegatecall, then extracts the Variable being passed in as the target.
    Can return a newly created StateVariable if an `sload` from a hardcoded storage slot is found in assembly.
    Should typically be called by get_proxy_implementation_var(proxy).
    Args:
        proxy: A Contract object (should have a fallback function).

    Returns:
        (`Variable`) | None : The variable being passed as the destination argument in a delegatecall in the fallback.
    """
    delegate: Variable | None = None
    fallback = proxy.fallback_function
    for node in fallback.all_nodes():
        for ir in node.irs:
            if isinstance(ir, LowLevelCall) and ir.function_name == "delegatecall":
                delegate = ir.destination
        if delegate is not None:
            break
        if (
            node.type == NodeType.ASSEMBLY
            and isinstance(node.inline_asm, str)
            and "delegatecall" in node.inline_asm
        ):
            delegate = extract_delegate_from_asm(proxy, node)
        elif node.type == NodeType.EXPRESSION:
            expression = node.expression
            if isinstance(expression, AssignmentOperation):
                expression = expression.expression_right
            if (
                isinstance(expression, CallExpression)
                and "delegatecall" in str(expression.called)
                and len(expression.arguments) > 1
            ):
                dest = expression.arguments[1]
                if isinstance(dest, CallExpression) and "sload" in str(dest.called):
                    dest = dest.arguments[0]
                if isinstance(dest, Identifier):
                    delegate = dest.value
                    break
                if (
                    isinstance(dest, Literal) and len(dest.value) == 66
                ):  # 32 bytes = 64 chars + "0x" = 66 chars
                    # Storage slot is not declared as a constant, but rather is hardcoded in the assembly,
                    # so create a new StateVariable to represent it.
                    delegate = create_state_variable_from_slot(dest.value)
                    break
    return delegate


def extract_delegate_from_asm(contract: Contract, node: Node) -> Variable | None:
    """
    Finds a Variable with a name matching the argument passed into a delegatecall, when all we have is an Assembly node
    with a block of code as one long string. Usually only the case for solc versions < 0.6.0.
    Can return a newly created StateVariable if an `sload` from a hardcoded storage slot is found in assembly.
    Should typically be called by find_delegate_in_fallback(proxy).
    Args:
        contract: The parent Contract.
        node: The Assembly Node (i.e., node.type == NodeType.ASSEMBLY)

    Returns:
        (`Variable`) | None : The variable being passed as the destination argument in a delegatecall in the fallback.
    """
    asm_split = str(node.inline_asm).split("\n")
    asm = next(line for line in asm_split if "delegatecall" in line)
    params = asm.split("call(")[1].split(", ")
    dest = params[1]
    if dest.endswith(")") and not dest.startswith("sload("):
        dest = params[2]
    if dest.startswith("sload("):
        dest = dest.replace(")", "(").split("(")[1]
        if dest.startswith("0x"):
            return create_state_variable_from_slot(dest)
        if dest.isnumeric():
            slot_idx = int(dest)
            return next(
                (
                    v
                    for v in contract.state_variables_ordered
                    if SlitherReadStorage.get_variable_info(contract, v)[0] == slot_idx
                ),
                None,
            )
        for v in node.function.variables_read_or_written:
            if v.name == dest:
                if isinstance(v, LocalVariable) and v.expression is not None:
                    e = v.expression
                    if isinstance(e, Identifier) and isinstance(e.value, StateVariable):
                        v = e.value
                        # Fall through, return constant storage slot
                if isinstance(v, StateVariable) and v.is_constant:
                    return v
    if "_fallback_asm" in dest or "_slot" in dest:
        dest = dest.split("_")[0]
    return find_delegate_from_name(contract, dest, node.function)


def find_delegate_from_name(
    contract: Contract, dest: str, parent_func: Function
) -> Variable | None:
    """
    Searches for a variable with a given name, starting with StateVariables declared in the contract, followed by
    LocalVariables in the parent function, either declared in the function body or as parameters in the signature.
    Can return a newly created StateVariable if an `sload` from a hardcoded storage slot is found in assembly.
    Args:
        contract: The Contract object to search.
        dest: The variable name to search for.
        parent_func: The Function object to search.

    Returns:
        (`Variable`) | None : The variable with the matching name, if found
    """
    for sv in contract.state_variables:
        if sv.name == dest:
            return sv
    for lv in parent_func.local_variables:
        if lv.name == dest:
            return lv
    for pv in parent_func.parameters + parent_func.returns:
        if pv.name == dest:
            return pv
    if parent_func.contains_assembly:
        for node in parent_func.all_nodes():
            if node.type == NodeType.ASSEMBLY and isinstance(node.inline_asm, str):
                asm = next(
                    (
                        s
                        for s in node.inline_asm.split("\n")
                        if f"{dest}:=sload(" in s.replace(" ", "")
                    ),
                    None,
                )
                if asm:
                    slot = asm.split("sload(")[1].split(")")[0]
                    if slot.startswith("0x"):
                        return create_state_variable_from_slot(slot, name=dest)
                    try:
                        slot_idx = int(slot)
                        return next(
                            (
                                v
                                for v in contract.state_variables_ordered
                                if SlitherReadStorage.get_variable_info(contract, v)[0] == slot_idx
                            ),
                            None,
                        )
                    except TypeError:
                        continue
    return None


def create_state_variable_from_slot(slot: str, name: str | None = None) -> StateVariable | None:
    """
    Creates a new StateVariable object to wrap a hardcoded storage slot found in assembly.
    Args:
        slot: The storage slot hex string.
        name: Optional name for the variable. The slot string is used if name is not provided.

    Returns:
        A newly created constant StateVariable of type bytes32, with the slot as the variable's expression and name,
        if slot matches the length and prefix of a bytes32. Otherwise, returns None.
    """
    if len(slot) == 66 and slot.startswith("0x"):  # 32 bytes = 64 chars + "0x" = 66 chars
        # Storage slot is not declared as a constant, but rather is hardcoded in the assembly,
        # so create a new StateVariable to represent it.
        v = StateVariable()
        v.is_constant = True
        v.expression = Literal(slot, ElementaryType("bytes32"))
        if name is not None:
            v.name = name
        else:
            v.name = slot
        v.type = ElementaryType("bytes32")
        return v
    # This should probably also handle hashed strings, but for now return None
    return None



"""
Module detecting misuse of Boolean constants
"""

from slither.core.cfg.node import Node, NodeType
from slither.core.declarations import Function
from slither.core.declarations.contract import Contract
from slither.core.solidity_types import ElementaryType
from slither.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DETECTOR_INFO,
)
from slither.slithir.operations import (
    Assignment,
    Call,
    Return,
    InitArray,
    Binary,
    BinaryType,
    Condition,
)
from slither.slithir.variables import Constant
from slither.utils.output import Output


class BooleanConstantMisuse(AbstractDetector):
    """
    Boolean constant misuse
    """

    ARGUMENT = "boolean-cst"
    HELP = "Misuse of Boolean constant"
    IMPACT = DetectorClassification.MEDIUM
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#misuse-of-a-boolean-constant"
    )

    WIKI_TITLE = "Misuse of a Boolean constant"
    WIKI_DESCRIPTION = """Detects the misuse of a Boolean constant."""

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract A {
	function f(uint x) public {
		// ...
        if (false) { // bad!
           // ...
        }
		// ...
	}

	function g(bool b) public returns (bool) {
		// ...
        return (b || true); // bad!
		// ...
	}
}
```
Boolean constants in code have only a few legitimate uses.
Other uses (in complex expressions, as conditionals) indicate either an error or, most likely, the persistence of faulty code."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = """Verify and simplify the condition."""

    @staticmethod
    def _detect_boolean_constant_misuses(
        contract: Contract,
    ) -> list[tuple[Function, set[Node]]]:
        """
        Detects and returns all nodes which misuse a Boolean constant.
        :param contract: Contract to detect assignment within.
        :return: A list of misusing nodes.
        """

        # Create our result set.
        results: list[tuple[Function, set[Node]]] = []

        # Loop for each function and modifier.
        for function in contract.functions_declared:
            f_results = set()

            # Loop for every node in this function, looking for boolean constants
            for node in function.nodes:
                # Do not report "while(true)"
                if node.type == NodeType.IFLOOP and node.irs and len(node.irs) == 1:
                    ir = node.irs[0]
                    if isinstance(ir, Condition) and ir.value == Constant(
                        "True", ElementaryType("bool")
                    ):
                        continue

                for ir in node.irs:
                    if isinstance(ir, (Assignment, Call, Return, InitArray)):
                        # It's ok to use a bare boolean constant in these contexts
                        continue
                    if isinstance(ir, Binary) and ir.type in [
                        BinaryType.ADDITION,
                        BinaryType.EQUAL,
                        BinaryType.NOT_EQUAL,
                    ]:
                        # Comparing to a Boolean constant is dubious style, but harmless
                        # Equal is catch by another detector (informational severity)
                        continue
                    for r in ir.read:
                        if isinstance(r, Constant) and isinstance(r.value, bool):
                            f_results.add(node)
                results.append((function, f_results))

        # Return the resulting set of nodes with improper uses of Boolean constants
        return results

    def _detect(self) -> list[Output]:
        """
        Detect Boolean constant misuses
        """
        results = []
        for contract in self.contracts:
            boolean_constant_misuses = self._detect_boolean_constant_misuses(contract)
            for func, nodes in boolean_constant_misuses:
                for node in nodes:
                    info: DETECTOR_INFO = [
                        func,
                        " uses a Boolean constant improperly:\n\t-",
                        node,
                        "\n",
                    ]

                    res = self.generate_result(info)
                    results.append(res)

        return results



from slither.core.expressions.unary_operation import UnaryOperationType, UnaryOperation
from slither.tools.mutator.utils.patch import create_patch_with_line
from slither.tools.mutator.mutators.abstract_mutator import AbstractMutator

unary_operators = [
    UnaryOperationType.PLUSPLUS_PRE,
    UnaryOperationType.MINUSMINUS_PRE,
    UnaryOperationType.PLUSPLUS_POST,
    UnaryOperationType.MINUSMINUS_POST,
    UnaryOperationType.MINUS_PRE,
]


class UOR(AbstractMutator):
    NAME = "UOR"
    HELP = "Unary Operator Replacement"

    def _mutate(self) -> dict:
        result: dict = {}

        for function in self.contract.functions_and_modifiers_declared:
            if not self.should_mutate_function(function):
                continue
            for node in function.nodes:
                if not self.should_mutate_node(node):
                    continue
                try:
                    ir_expression = node.expression
                except AttributeError:
                    continue
                start = node.source_mapping.start
                stop = start + node.source_mapping.length
                old_str = node.source_mapping.content
                line_no = node.source_mapping.lines
                if (
                    isinstance(ir_expression, UnaryOperation)
                    and ir_expression.type in unary_operators
                ):
                    for op in unary_operators:
                        if not node.expression.is_prefix:
                            if node.expression.type != op:
                                variable_read = node.variables_read[0]
                                new_str = str(variable_read) + str(op)
                                if new_str != old_str and str(op) != "-":
                                    create_patch_with_line(
                                        result,
                                        self.in_file,
                                        start,
                                        stop,
                                        old_str,
                                        new_str,
                                        line_no[0],
                                    )
                                new_str = str(op) + str(variable_read)
                                create_patch_with_line(
                                    result,
                                    self.in_file,
                                    start,
                                    stop,
                                    old_str,
                                    new_str,
                                    line_no[0],
                                )
                        else:
                            if node.expression.type != op:
                                variable_read = node.variables_read[0]
                                new_str = str(op) + str(variable_read)
                                if new_str != old_str and str(op) != "-":
                                    create_patch_with_line(
                                        result,
                                        self.in_file,
                                        start,
                                        stop,
                                        old_str,
                                        new_str,
                                        line_no[0],
                                    )
                                new_str = str(variable_read) + str(op)
                                create_patch_with_line(
                                    result,
                                    self.in_file,
                                    start,
                                    stop,
                                    old_str,
                                    new_str,
                                    line_no[0],
                                )
        return result



""" "
Re-entrancy detection

Based on heuristics, it may lead to FP and FN
Iterate over all the nodes of the graph until reaching a fixpoint
"""

from collections import namedtuple, defaultdict

from slither.detectors.abstract_detector import DetectorClassification
from slither.detectors.reentrancy.reentrancy import Reentrancy, to_hashable
from slither.utils.output import Output

FindingKey = namedtuple("FindingKey", ["function", "calls", "send_eth"])
FindingValue = namedtuple("FindingValue", ["variable", "node", "nodes"])


class ReentrancyEvent(Reentrancy):
    ARGUMENT = "reentrancy-events"
    HELP = "Reentrancy vulnerabilities leading to out-of-order Events"
    IMPACT = DetectorClassification.LOW
    CONFIDENCE = DetectorClassification.MEDIUM

    WIKI = (
        "https://github.com/crytic/slither/wiki/Detector-Documentation#reentrancy-vulnerabilities-4"
    )

    WIKI_TITLE = "Reentrancy vulnerabilities"

    # region wiki_description
    WIKI_DESCRIPTION = """
Detects [reentrancies](https://github.com/trailofbits/not-so-smart-contracts/tree/master/reentrancy) that allow manipulation of the order or value of events."""
    # endregion wiki_description

    # region wiki_exploit_scenario
    WIKI_EXPLOIT_SCENARIO = """
```solidity
contract ReentrantContract {
	function f() external {
		if (BugReentrancyEvents(msg.sender).counter() == 1) {
			BugReentrancyEvents(msg.sender).count(this);
		}
	}
}
contract Counter {
	uint public counter;
	event Counter(uint);

}
contract BugReentrancyEvents is Counter {
    function count(ReentrantContract d) external {
        counter += 1;
        d.f();
        emit Counter(counter);
    }
}
contract NoReentrancyEvents is Counter {
	function count(ReentrantContract d) external {
        counter += 1;
        emit Counter(counter);
        d.f();
    }
}
```

If the external call `d.f()` re-enters `BugReentrancyEvents`, the `Counter` events will be incorrect (`Counter(2)`, `Counter(2)`) whereas `NoReentrancyEvents` will correctly emit
(`Counter(1)`, `Counter(2)`). This may cause issues for offchain components that rely on the values of events e.g. checking for the amount deposited to a bridge."""
    # endregion wiki_exploit_scenario

    WIKI_RECOMMENDATION = "Apply the [`check-effects-interactions` pattern](https://docs.soliditylang.org/en/latest/security-considerations.html#re-entrancy)."

    STANDARD_JSON = False

    def find_reentrancies(self) -> defaultdict[FindingKey, set[FindingValue]]:
        result = defaultdict(set)
        for contract in self.contracts:
            for f in contract.functions_and_modifiers_declared:
                if not f.is_reentrant:
                    continue
                for node in f.nodes:
                    # dead code
                    if self.KEY not in node.context:
                        continue
                    if node.context[self.KEY].calls:
                        if not any(n != node for n in node.context[self.KEY].calls):
                            continue

                        # calls are ordered
                        finding_key = FindingKey(
                            function=node.function,
                            calls=to_hashable(node.context[self.KEY].calls),
                            send_eth=to_hashable(node.context[self.KEY].send_eth),
                        )
                        finding_vars = {
                            FindingValue(
                                e,
                                e.node,
                                tuple(sorted(nodes, key=lambda x: x.node_id)),
                            )
                            for (e, nodes) in node.context[self.KEY].events.items()
                        }
                        if finding_vars:
                            result[finding_key] |= finding_vars
        return result

    def _detect(self) -> list[Output]:
        """"""
        super()._detect()

        reentrancies = self.find_reentrancies()

        results = []

        result_sorted = sorted(reentrancies.items(), key=lambda x: x[0][0].name)
        for (func, calls, send_eth), events in result_sorted:
            calls = sorted(set(calls), key=lambda x: x[0].node_id)
            send_eth = sorted(set(send_eth), key=lambda x: x[0].node_id)
            events = sorted(events, key=lambda x: (str(x.variable.name), x.node.node_id))

            info = ["Reentrancy in ", func, ":\n"]
            info += ["\tExternal calls:\n"]
            for call_info, calls_list in calls:
                info += ["\t- ", call_info, "\n"]
                for call_list_info in calls_list:
                    if call_list_info != call_info:
                        info += ["\t\t- ", call_list_info, "\n"]
            if calls != send_eth and send_eth:
                info += ["\tExternal calls sending eth:\n"]
                for call_info, calls_list in send_eth:
                    info += ["\t- ", call_info, "\n"]
                    for call_list_info in calls_list:
                        if call_list_info != call_info:
                            info += ["\t\t- ", call_list_info, "\n"]
            info += ["\tEvent emitted after the call(s):\n"]
            for finding_value in events:
                info += ["\t- ", finding_value.node, "\n"]
                for other_node in finding_value.nodes:
                    if other_node != finding_value.node:
                        info += ["\t\t- ", other_node, "\n"]

            # Create our JSON result
            res = self.generate_result(info)

            # Add the function with the re-entrancy first
            res.add(func)

            # Add all underlying calls in the function which are potentially problematic.
            for call_info, calls_list in calls:
                res.add(call_info, {"underlying_type": "external_calls"})
                for call_list_info in calls_list:
                    if call_list_info != call_info:
                        res.add(
                            call_list_info,
                            {"underlying_type": "external_calls_sending_eth"},
                        )

            #

            # If the calls are not the same ones that send eth, add the eth sending nodes.
            if calls != send_eth:
                for call_info, calls_list in send_eth:
                    res.add(call_info, {"underlying_type": "external_calls_sending_eth"})
                    for call_list_info in calls_list:
                        if call_list_info != call_info:
                            res.add(
                                call_list_info,
                                {"underlying_type": "external_calls_sending_eth"},
                            )

            for finding_value in events:
                res.add(finding_value.node, {"underlying_type": "event"})
                for other_node in finding_value.nodes:
                    if other_node != finding_value.node:
                        res.add(other_node, {"underlying_type": "event"})

            # Append our result
            results.append(res)

        return results



import logging
from math import floor
from typing import Any
from collections.abc import Callable

import dataclasses

from eth_abi import decode, encode
from eth_typing.evm import ChecksumAddress
from eth_utils import keccak, to_checksum_address
from web3 import Web3
from web3.types import BlockIdentifier
from web3.exceptions import ExtraDataLengthError
from web3.middleware import ExtraDataToPOAMiddleware

from slither.core.declarations import Contract, Structure
from slither.core.solidity_types import ArrayType, ElementaryType, MappingType, UserDefinedType
from slither.core.solidity_types.type import Type
from slither.core.cfg.node import NodeType
from slither.core.variables.state_variable import StateVariable
from slither.core.variables.structure_variable import StructureVariable
from slither.core.expressions import (
    AssignmentOperation,
    Literal,
    Identifier,
    BinaryOperation,
    UnaryOperation,
    TupleExpression,
    TypeConversion,
    CallExpression,
)
from slither.utils.myprettytable import MyPrettyTable
from slither.visitors.expression.constants_folding import ConstantFolding, NotConstant

from .utils import coerce_type, get_offset_value, get_storage_data

logging.basicConfig()
logger = logging.getLogger("Slither-read-storage")
logger.setLevel(logging.INFO)

Elem = dict[str, "SlotInfo"]
NestedElem = dict[str, Elem]


@dataclasses.dataclass(slots=True)
class SlotInfo:
    name: str
    type_string: str
    slot: int
    size: int
    offset: int
    value: int | bool | str | ChecksumAddress | None = None
    # For structure and array, str->SlotInfo
    elems: Elem | NestedElem = dataclasses.field(default_factory=lambda: {})  # type: ignore[assignment]


class SlitherReadStorageException(Exception):
    pass


class RpcInfo:
    def __init__(self, rpc_url: str, block: BlockIdentifier = "latest") -> None:
        assert isinstance(block, int) or block in [
            "latest",
            "earliest",
            "pending",
            "safe",
            "finalized",
        ]
        self.rpc: str = rpc_url
        self._web3: Web3 = Web3(Web3.HTTPProvider(self.rpc))
        """If the RPC is for a POA network, the first call to get_block fails, so we inject geth_poa_middleware"""
        try:
            self._block: int = self.web3.eth.get_block(block)["number"]
        except ExtraDataLengthError:
            self._web3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)
            self._block: int = self.web3.eth.get_block(block)["number"]

    @property
    def web3(self) -> Web3:
        return self._web3

    @property
    def block(self) -> int:
        return self._block


class SlitherReadStorage:
    def __init__(self, contracts: list[Contract], max_depth: int, rpc_info: RpcInfo = None) -> None:
        self._checksum_address: ChecksumAddress | None = None
        self._contracts: list[Contract] = contracts
        self._log: str = ""
        self._max_depth: int = max_depth
        self._slot_info: dict[str, SlotInfo] = {}
        self._target_variables: list[tuple[Contract, StateVariable]] = []
        self._constant_storage_slots: list[tuple[Contract, StateVariable]] = []
        self.rpc_info: RpcInfo | None = rpc_info
        self.storage_address: str | None = None
        self.table: MyPrettyTable | None = None
        self.unstructured: bool = False

    @property
    def contracts(self) -> list[Contract]:
        return self._contracts

    @property
    def max_depth(self) -> int:
        return int(self._max_depth)

    @property
    def log(self) -> str:
        return self._log

    @log.setter
    def log(self, log: str) -> None:
        self._log = log

    @property
    def checksum_address(self) -> ChecksumAddress:
        if not self.storage_address:
            raise ValueError
        if not self._checksum_address:
            self._checksum_address = to_checksum_address(self.storage_address)
        return self._checksum_address

    @property
    def target_variables(self) -> list[tuple[Contract, StateVariable]]:
        """Storage variables (not constant or immutable) and their associated contract."""
        return self._target_variables

    @property
    def constant_slots(self) -> list[tuple[Contract, StateVariable]]:
        """Constant bytes32 variables and their associated contract."""
        return self._constant_storage_slots

    @property
    def slot_info(self) -> dict[str, SlotInfo]:
        """Contains the location, type, size, offset, and value of contract slots."""
        return self._slot_info

    def get_storage_layout(self) -> None:
        """Retrieves the storage layout of entire contract."""
        tmp: dict[str, SlotInfo] = {}
        for contract, var in self.target_variables:
            type_ = var.type
            info = self.get_storage_slot(var, contract)
            if info:
                tmp[var.name] = info
                if isinstance(type_, UserDefinedType) and isinstance(type_.type, Structure):
                    tmp[var.name].elems = self._all_struct_slots(var, type_.type, contract)

                elif isinstance(type_, ArrayType):
                    elems = self._all_array_slots(var, contract, type_, info.slot)
                    tmp[var.name].elems = elems
        if self.unstructured:
            tmp.update(self.get_unstructured_layout())
        self._slot_info = tmp

    def get_unstructured_layout(self) -> dict[str, SlotInfo]:
        tmp: dict[str, SlotInfo] = {}
        for _, var in self.constant_slots:
            var_name = var.name
            try:
                exp = var.expression
                if isinstance(
                    exp,
                    (
                        BinaryOperation,
                        UnaryOperation,
                        Identifier,
                        TupleExpression,
                        TypeConversion,
                        CallExpression,
                    ),
                ):
                    exp = ConstantFolding(exp, "bytes32").result()
                if isinstance(exp, Literal):
                    slot = coerce_type("int", exp.value)
                else:
                    continue
                offset = 0
                type_string, size = self.find_constant_slot_storage_type(var)
                if type_string:
                    tmp[var.name] = SlotInfo(
                        name=var_name, type_string=type_string, slot=slot, size=size, offset=offset
                    )
                    self.log += (
                        f"\nSlot Name: {var_name}\nType: bytes32"
                        f"\nStorage Type: {type_string}\nSlot: {exp!s}\n"
                    )
                    logger.info(self.log)
                    self.log = ""
            except NotConstant:
                continue
        return tmp

    def get_storage_slot(
        self,
        target_variable: StateVariable,
        contract: Contract,
        **kwargs: Any,
    ) -> SlotInfo | None:
        """
        Finds the storage slot of a variable in a given contract.
        Args:
            target_variable (`StateVariable`): The variable to retrieve the slot for.
            contracts (`Contract`): The contract that contains the given state variable.
        **kwargs:
            key (int): Key of a mapping or index position if an array.
            deep_key (int): Key of a mapping embedded within another mapping or
            secondary index if array.
            struct_var (str): Structure variable name.
        Returns:
            (`SlotInfo`) | None : A dictionary of the slot information.
        """

        key: int | None = kwargs.get("key")
        deep_key: int | None = kwargs.get("deep_key")
        struct_var: str | None = kwargs.get("struct_var")
        info: str
        var_log_name = target_variable.name
        try:
            int_slot, size, offset, type_to = self.get_variable_info(contract, target_variable)
        except KeyError:
            # Only the child contract of a parent contract will show up in the storage layout when inheritance is used
            logger.info(
                f"\nContract {contract} not found in storage layout. It is possibly a parent contract\n"
            )
            return None

        slot = int.to_bytes(int_slot, 32, byteorder="big")

        target_variable_type = target_variable.type

        if isinstance(target_variable_type, ElementaryType):
            type_to = target_variable_type.name

        elif isinstance(target_variable_type, ArrayType) and key is not None:
            var_log_name = f"{var_log_name}[{key}]"
            info, type_to, slot, size, offset = self._find_array_slot(
                target_variable_type,
                slot,
                key,
                deep_key=deep_key,
                struct_var=struct_var,
            )
            self.log += info

        elif isinstance(target_variable_type, UserDefinedType) and struct_var is not None:
            var_log_name = f"{var_log_name}.{struct_var}"
            target_variable_type_type = target_variable_type.type
            assert isinstance(target_variable_type_type, Structure)
            elems = target_variable_type_type.elems_ordered
            info, type_to, slot, size, offset = self._find_struct_var_slot(elems, slot, struct_var)
            self.log += info

        elif isinstance(target_variable_type, MappingType) and key:
            info, type_to, slot, size, offset = self._find_mapping_slot(
                target_variable_type, slot, key, struct_var=struct_var, deep_key=deep_key
            )
            self.log += info

        int_slot = int.from_bytes(slot, byteorder="big")
        self.log += f"\nName: {var_log_name}\nType: {type_to}\nSlot: {int_slot}\n"
        logger.info(self.log)
        self.log = ""

        return SlotInfo(
            name=var_log_name,
            type_string=type_to,
            slot=int_slot,
            size=size,
            offset=offset,
        )

    def get_target_variables(self, **kwargs) -> None:
        """
        Retrieves every instance of a given variable in a list of contracts.
        Should be called after setting `target_variables` with `get_all_storage_variables()`.
        **kwargs:
            key (str): Key of a mapping or index position if an array.
            deep_key (str): Key of a mapping embedded within another mapping or secondary index if array.
            struct_var (str): Structure variable name.
        """
        for contract, var in self.target_variables:
            slot_info = self.get_storage_slot(var, contract, **kwargs)
            if slot_info:
                self._slot_info[f"{contract.name}.{var.name}"] = slot_info

    def find_constant_slot_storage_type(self, var: StateVariable) -> tuple[str | None, int | None]:
        """
        Given a constant bytes32 StateVariable, tries to determine which variable type is stored there, using the
        heuristic that if a function reads from the slot and returns a value, it probably stores that type of value.
        Also uses the StorageSlot library as a heuristic when a function has no return but uses the library's getters.
        Args:
            var (StateVariable): The constant bytes32 storage slot.

        Returns:
            type (str): The type of value stored in the slot.
            size (int): The type's size in bits.
        """
        assert var.is_constant and var.type == ElementaryType("bytes32")
        storage_type = None
        size = None
        funcs = []
        for c in self.contracts:
            c_funcs = c.get_functions_reading_from_variable(var)
            c_funcs.extend(
                f
                for f in c.functions
                if any(str(v.expression) == str(var.expression) for v in f.variables)
            )
            c_funcs = list(set(c_funcs))
            funcs.extend(c_funcs)
        fallback = [f for f in var.contract.functions if f.is_fallback]
        funcs += fallback
        for func in funcs:
            rets = func.return_type if func.return_type is not None else []
            for ret in rets:
                size, _ = ret.storage_size
                if size <= 32:
                    return str(ret), size * 8
            for node in func.all_nodes():
                exp = node.expression
                # Look for use of the common OpenZeppelin StorageSlot library
                if f"getAddressSlot({var.name})" in str(exp):
                    return "address", 160
                if f"getBooleanSlot({var.name})" in str(exp):
                    return "bool", 1
                if f"getBytes32Slot({var.name})" in str(exp):
                    return "bytes32", 256
                if f"getUint256Slot({var.name})" in str(exp):
                    return "uint256", 256
                # Look for variable assignment in assembly loaded from a hardcoded slot
                if (
                    isinstance(exp, AssignmentOperation)
                    and isinstance(exp.expression_left, Identifier)
                    and isinstance(exp.expression_right, CallExpression)
                    and "sload" in str(exp.expression_right.called)
                    and str(exp.expression_right.arguments[0]) == str(var.expression)
                ):
                    if func.is_fallback:
                        return "address", 160
                    storage_type = exp.expression_left.value.type.name
                    size, _ = exp.expression_left.value.type.storage_size
                    return storage_type, size * 8
                # Look for variable storage in assembly stored to a hardcoded slot
                if (
                    isinstance(exp, CallExpression)
                    and "sstore" in str(exp.called)
                    and isinstance(exp.arguments[0], Identifier)
                    and isinstance(exp.arguments[1], Identifier)
                    and str(exp.arguments[0].value.expression) == str(var.expression)
                ):
                    storage_type = exp.arguments[1].value.type.name
                    size, _ = exp.arguments[1].value.type.storage_size
                    return storage_type, size * 8
        return storage_type, size

    def walk_slot_info(self, func: Callable) -> None:
        stack = list(self.slot_info.values())
        while stack:
            slot_info = stack.pop()
            if isinstance(slot_info, dict):  # NestedElem
                stack.extend(slot_info.values())
            elif slot_info.elems:
                stack.extend(list(slot_info.elems.values()))
            if isinstance(slot_info, SlotInfo):
                func(slot_info)

    def get_slot_values(self, slot_info: SlotInfo) -> None:
        """
        Fetches the slot value of `SlotInfo` object
        :param slot_info:
        """
        assert self.rpc_info is not None
        hex_bytes = get_storage_data(
            self.rpc_info.web3,
            self.checksum_address,
            int.to_bytes(slot_info.slot, 32, byteorder="big"),
            self.rpc_info.block,
        )
        slot_info.value = self.convert_value_to_type(
            hex_bytes, slot_info.size, slot_info.offset, slot_info.type_string
        )
        logger.info(f"\nValue: {slot_info.value}\n")

    def get_all_storage_variables(self, func: Callable = lambda x: x) -> None:
        """
        Fetches all storage variables from a list of contracts.
        kwargs:
            func (Callable, optional): A criteria to filter functions e.g. name.
        """
        for contract in self.contracts:
            for var in contract.state_variables_ordered:
                if func(var):
                    if var.is_stored:
                        self._target_variables.append((contract, var))
                    elif (
                        self.unstructured
                        and var.is_constant
                        and var.type == ElementaryType("bytes32")
                    ):
                        self._constant_storage_slots.append((contract, var))
            if self.unstructured:
                hardcoded_slot = self.find_hardcoded_slot_in_fallback(contract)
                if hardcoded_slot is not None:
                    self._constant_storage_slots.append((contract, hardcoded_slot))

    def find_hardcoded_slot_in_fallback(self, contract: Contract) -> StateVariable | None:
        """
        Searches the contract's fallback function for a sload from a literal storage slot, i.e.,
        `let contractLogic := sload(0xc5f16f0fcc639fa48a6947836d9850f504798523bf8c9a3a87d5876cf622bcf7)`.

        Args:
            contract: a Contract object, which should have a fallback function.

        Returns:
            A newly created StateVariable representing the Literal bytes32 slot, if one is found, otherwise None.
        """
        fallback = None
        for func in contract.functions_entry_points:
            if func.is_fallback:
                fallback = func
                break
        if fallback is None:
            return None
        queue = [fallback.entry_point]
        visited = []
        while len(queue) > 0:
            node = queue.pop(0)
            visited.append(node)
            queue.extend(son for son in node.sons if son not in visited)
            if node.type == NodeType.ASSEMBLY and isinstance(node.inline_asm, str):
                return SlitherReadStorage.find_hardcoded_slot_in_asm_str(node.inline_asm, contract)
            if node.type == NodeType.EXPRESSION:
                sv = self.find_hardcoded_slot_in_exp(node.expression, contract)
                if sv is not None:
                    return sv
        return None

    @staticmethod
    def find_hardcoded_slot_in_asm_str(inline_asm: str, contract: Contract) -> StateVariable | None:
        """
        Searches a block of assembly code (given as a string) for a sload from a literal storage slot.
        Does not work if the argument passed to sload does not start with "0x", i.e., `sload(add(1,1))`
        or `and(sload(0), 0xffffffffffffffffffffffffffffffffffffffff)`.

        Args:
            inline_asm: a string containing all the code in an assembly node (node.inline_asm for solc < 0.6.0).

        Returns:
            A newly created StateVariable representing the Literal bytes32 slot, if one is found, otherwise None.
        """
        asm_split = inline_asm.split("\n")
        for asm in asm_split:
            if "sload(" in asm:  # Only handle literals
                arg = asm.split("sload(")[1].split(")")[0]
                if arg.startswith("0x"):
                    exp = Literal(arg, ElementaryType("bytes32"))
                    sv = StateVariable()
                    sv.name = "fallback_sload_hardcoded"
                    sv.expression = exp
                    sv.is_constant = True
                    sv.type = exp.type
                    sv.set_contract(contract)
                    return sv
        return None

    def find_hardcoded_slot_in_exp(
        self, exp: "Expression", contract: Contract
    ) -> StateVariable | None:
        """
        Parses an expression to see if it contains a sload from a literal storage slot,
        unrolling nested expressions if necessary to determine which slot it loads from.
        Args:
            exp: an Expression object to search.
            contract: the Contract containing exp.

        Returns:
            A newly created StateVariable representing the Literal bytes32 slot, if one is found, otherwise None.
        """
        if isinstance(exp, AssignmentOperation):
            exp = exp.expression_right
        while isinstance(exp, BinaryOperation):
            exp = next(
                (e for e in exp.expressions if isinstance(e, (CallExpression, BinaryOperation))),
                exp.expression_left,
            )
        while isinstance(exp, CallExpression) and len(exp.arguments) > 0:
            called = exp.called
            exp = exp.arguments[0]
            if "sload" in str(called):
                break
        if isinstance(
            exp,
            (
                BinaryOperation,
                UnaryOperation,
                Identifier,
                TupleExpression,
                TypeConversion,
                CallExpression,
            ),
        ):
            try:
                exp = ConstantFolding(exp, "bytes32").result()
            except NotConstant:
                return None
        if (
            isinstance(exp, Literal)
            and isinstance(exp.type, ElementaryType)
            and exp.type.name in ["bytes32", "uint256"]
        ):
            sv = StateVariable()
            sv.name = "fallback_sload_hardcoded"
            value = exp.value
            str_value = str(value)
            if str_value.isdecimal():
                value = int(value)
            if isinstance(value, (int, bytes)):
                if isinstance(value, bytes):
                    str_value = "0x" + value.hex()
                    value = int(str_value, 16)
                exp = Literal(str_value, ElementaryType("bytes32"))
                state_var_slots = [
                    self.get_variable_info(contract, var)[0]
                    for contract, var in self.target_variables
                ]
                if value in state_var_slots:
                    return None
            sv.expression = exp
            sv.is_constant = True
            sv.type = ElementaryType("bytes32")
            sv.set_contract(contract)
            return sv
        return None

    def convert_slot_info_to_rows(self, slot_info: SlotInfo) -> None:
        """
        Convert and append slot info to table. Create table if it
        does not yet exist
        :param slot_info:
        """
        field_names = [
            field.name for field in dataclasses.fields(SlotInfo) if field.name != "elems"
        ]
        if not self.table:
            self.table = MyPrettyTable(field_names)
        self.table.add_row([getattr(slot_info, field) for field in field_names])

    def to_json(self) -> dict:
        return {key: dataclasses.asdict(value) for key, value in self.slot_info.items()}

    @staticmethod
    def _find_struct_var_slot(
        elems: list[StructureVariable], slot_as_bytes: bytes, struct_var: str
    ) -> tuple[str, str, bytes, int, int]:
        """
        Finds the slot of a structure variable.
        Args:
            elems (List[StructureVariable]): Ordered list of structure variables.
            slot_as_bytes (bytes): The slot of the struct to begin searching at.
            struct_var (str): The target structure variable.
        Returns:
            info (str): Info about the target variable to log.
            type_to (str): The type of the target variable.
            slot (bytes): The storage location of the target variable.
            size (int): The size (in bits) of the target variable.
            offset (int): The size of other variables that share the same slot.
        """
        slot = int.from_bytes(slot_as_bytes, "big")
        offset = 0
        type_to = ""
        size = 0
        for var in elems:
            var_type = var.type
            if isinstance(var_type, ElementaryType):
                size = var_type.size
                if size > (256 - offset):
                    slot += 1
                    offset = 0
                if struct_var == var.name:
                    type_to = var_type.name
                    break  # found struct var
                offset += size
            else:
                logger.info(f"{type(var_type)} is current not implemented in _find_struct_var_slot")

        slot_as_bytes = int.to_bytes(slot, 32, byteorder="big")
        info = f"\nStruct Variable: {struct_var}"
        return info, type_to, slot_as_bytes, size, offset

    @staticmethod
    def _find_array_slot(
        target_variable_type: ArrayType,
        slot: bytes,
        key: int,
        deep_key: int | None = None,
        struct_var: str | None = None,
    ) -> tuple[str, str, bytes, int, int]:
        """
        Finds the slot of array's index.
        Args:
            target_variable (`StateVariable`): The array that contains the target variable.
            slot (bytes): The starting slot of the array.
            key (int): The target variable's index position.
            deep_key (int, optional): Secondary index if nested array.
            struct_var (str, optional): Structure variable name.
        Returns:
            info (str): Info about the target variable to log.
            type_to (str): The type of the target variable.
            slot (bytes): The storage location of the target variable.
            size (int): The size
            offset (int): The offset
        """
        info = f"\nKey: {key}"
        offset = 0
        size = 256

        target_variable_type_type = target_variable_type.type

        if isinstance(
            target_variable_type_type, ArrayType
        ):  # multidimensional array uint[i][], , uint[][i], or uint[][]
            assert isinstance(target_variable_type_type.type, ElementaryType)
            size = target_variable_type_type.type.size
            type_to = target_variable_type_type.type.name

            if target_variable_type.is_fixed_array:  # uint[][i]
                slot_int = int.from_bytes(slot, "big") + int(key)
            else:
                slot = keccak(slot)
                key = int(key)
                if target_variable_type_type.is_fixed_array:  # arr[i][]
                    key *= int(str(target_variable_type_type.length))
                slot_int = int.from_bytes(slot, "big") + key

            if not deep_key:
                return info, type_to, int.to_bytes(slot_int, 32, "big"), size, offset

            info += f"\nDeep Key: {deep_key}"
            if target_variable_type_type.is_dynamic_array:  # uint[][]
                # keccak256(keccak256(slot) + index) + floor(j / floor(256 / size))
                slot = keccak(int.to_bytes(slot_int, 32, "big"))
                slot_int = int.from_bytes(slot, "big")

            # keccak256(slot) + index + floor(j / floor(256 / size))
            slot_int += floor(int(deep_key) / floor(256 / size))  # uint[i][]

        elif target_variable_type.is_fixed_array:
            slot_int = int.from_bytes(slot, "big") + int(key)
            if isinstance(target_variable_type_type, UserDefinedType) and isinstance(
                target_variable_type_type.type, Structure
            ):  # struct[i]
                type_to = target_variable_type_type.type.name
                if not struct_var:
                    return info, type_to, int.to_bytes(slot_int, 32, "big"), size, offset
                elems = target_variable_type_type.type.elems_ordered
                slot = int.to_bytes(slot_int, 32, byteorder="big")
                info_tmp, type_to, slot, size, offset = SlitherReadStorage._find_struct_var_slot(
                    elems, slot, struct_var
                )
                info += info_tmp

            else:
                assert isinstance(target_variable_type_type, ElementaryType)
                type_to = target_variable_type_type.name
                size = target_variable_type_type.size  # bits

        elif isinstance(target_variable_type_type, UserDefinedType) and isinstance(
            target_variable_type_type.type, Structure
        ):  # struct[]
            slot = keccak(slot)
            slot_int = int.from_bytes(slot, "big") + int(key)
            type_to = target_variable_type_type.type.name
            if not struct_var:
                return info, type_to, int.to_bytes(slot_int, 32, "big"), size, offset
            elems = target_variable_type_type.type.elems_ordered
            slot = int.to_bytes(slot_int, 32, byteorder="big")
            info_tmp, type_to, slot, size, offset = SlitherReadStorage._find_struct_var_slot(
                elems, slot, struct_var
            )
            info += info_tmp

        else:
            assert isinstance(target_variable_type_type, ElementaryType)

            slot = keccak(slot)
            slot_int = int.from_bytes(slot, "big") + int(key)
            type_to = target_variable_type_type.name
            size = target_variable_type_type.size  # bits

        slot = int.to_bytes(slot_int, 32, byteorder="big")

        return info, type_to, slot, size, offset

    @staticmethod
    def _find_mapping_slot(
        target_variable_type: MappingType,
        slot: bytes,
        key: int | str,
        deep_key: int | str | None = None,
        struct_var: str | None = None,
    ) -> tuple[str, str, bytes, int, int]:
        """
        Finds the data slot of a target variable within a mapping.
            target_variable (`StateVariable`): The mapping that contains the target variable.
            slot (bytes): The starting slot of the mapping.
            key (Union[int, str]): The key the variable is stored at.
            deep_key (int, optional): Key of a mapping embedded within another mapping.
            struct_var (str, optional): Structure variable name.
        :returns:
            log (str): Info about the target variable to log.
            type_to (str): The type of the target variable.
            slot (bytes): The storage location of the target variable.
            size (int): The size (in bits) of the target variable.
            offset (int): The size of other variables that share the same slot.

        """
        info = ""
        offset = 0
        if key:
            info += f"\nKey: {key}"
        if deep_key:
            info += f"\nDeep Key: {deep_key}"
        assert isinstance(target_variable_type.type_from, ElementaryType)
        key_type = target_variable_type.type_from.name
        assert key
        if "int" in key_type:  # without this eth_utils encoding fails
            key = int(key)
        key = coerce_type(key_type, key)
        slot = keccak(encode([key_type, "uint256"], [key, decode(["uint256"], slot)[0]]))

        if isinstance(target_variable_type.type_to, UserDefinedType) and isinstance(
            target_variable_type.type_to.type, Structure
        ):  # mapping(elem => struct)
            assert struct_var
            elems = target_variable_type.type_to.type.elems_ordered
            info_tmp, type_to, slot, size, offset = SlitherReadStorage._find_struct_var_slot(
                elems, slot, struct_var
            )
            info += info_tmp

        elif isinstance(
            target_variable_type.type_to, MappingType
        ):  # mapping(elem => mapping(elem => ???))
            assert deep_key
            assert isinstance(target_variable_type.type_to.type_from, ElementaryType)
            key_type = target_variable_type.type_to.type_from.name
            if "int" in key_type:  # without this eth_utils encoding fails
                deep_key = int(deep_key)

            # If deep map, will be keccak256(abi.encode(key1, keccak256(abi.encode(key0, uint(slot)))))
            slot = keccak(encode([key_type, "bytes32"], [deep_key, slot]))

            # mapping(elem => mapping(elem => elem))
            target_variable_type_type_to_type_to = target_variable_type.type_to.type_to
            assert isinstance(
                target_variable_type_type_to_type_to, (UserDefinedType, ElementaryType)
            )
            type_to = str(target_variable_type_type_to_type_to.type)
            byte_size, _ = target_variable_type_type_to_type_to.storage_size
            size = byte_size * 8  # bits
            offset = 0

            if isinstance(target_variable_type_type_to_type_to, UserDefinedType) and isinstance(
                target_variable_type_type_to_type_to.type, Structure
            ):  # mapping(elem => mapping(elem => struct))
                assert struct_var
                elems = target_variable_type_type_to_type_to.type.elems_ordered
                # If map struct, will be bytes32(uint256(keccak256(abi.encode(key1, keccak256(abi.encode(key0, uint(slot)))))) + structFieldDepth);
                info_tmp, type_to, slot, size, offset = SlitherReadStorage._find_struct_var_slot(
                    elems, slot, struct_var
                )
                info += info_tmp

        # TODO: support mapping with dynamic arrays

        # mapping(elem => elem)
        elif isinstance(target_variable_type.type_to, ElementaryType):
            type_to = target_variable_type.type_to.name  # the value's elementary type
            byte_size, _ = target_variable_type.type_to.storage_size
            size = byte_size * 8  # bits

        else:
            raise NotImplementedError(
                f"{target_variable_type} => {target_variable_type.type_to} not implemented"
            )

        return info, type_to, slot, size, offset

    @staticmethod
    def get_variable_info(
        contract: Contract, target_variable: StateVariable
    ) -> tuple[int, int, int, str]:
        """Return slot, size, offset, and type."""
        assert isinstance(target_variable.type, Type)
        type_to = str(target_variable.type)
        byte_size, _ = target_variable.type.storage_size
        size = byte_size * 8  # bits
        (int_slot, offset) = contract.compilation_unit.storage_layout_of(contract, target_variable)
        offset *= 8  # bits
        logger.info(
            f"\nContract '{contract.name}'\n{target_variable.canonical_name} with type {target_variable.type} is located at slot: {int_slot}\n"
        )

        return int_slot, size, offset, type_to

    @staticmethod
    def convert_value_to_type(
        hex_bytes: bytes, size: int, offset: int, type_to: str
    ) -> int | bool | str | ChecksumAddress:
        """Convert slot data to type representation."""
        # Account for storage packing
        offset_hex_bytes = get_offset_value(hex_bytes, offset, size)
        try:
            value = coerce_type(type_to, offset_hex_bytes)
        except ValueError:
            return coerce_type("int", offset_hex_bytes)

        return value

    def _all_struct_slots(
        self, var: StateVariable, st: Structure, contract: Contract, key: int | None = None
    ) -> Elem:
        """Retrieves all members of a struct."""
        struct_elems = st.elems_ordered
        data: Elem = {}
        for elem in struct_elems:
            info = self.get_storage_slot(
                var,
                contract,
                key=key,
                struct_var=elem.name,
            )
            if info:
                data[elem.name] = info

        return data

    def _all_array_slots(
        self, var: StateVariable, contract: Contract, type_: ArrayType, slot: int
    ) -> Elem | NestedElem:
        """Retrieves all members of an array."""
        array_length = self._get_array_length(type_, slot)
        target_variable_type = type_.type
        if isinstance(target_variable_type, UserDefinedType) and isinstance(
            target_variable_type.type, Structure
        ):
            nested_elems: NestedElem = {}
            for i in range(min(array_length, self.max_depth)):
                nested_elems[str(i)] = self._all_struct_slots(
                    var, target_variable_type.type, contract, key=i
                )
            return nested_elems

        elems: Elem = {}
        for i in range(min(array_length, self.max_depth)):
            info = self.get_storage_slot(
                var,
                contract,
                key=str(i),
            )
            if info:
                elems[str(i)] = info

                if isinstance(target_variable_type, ArrayType):  # multidimensional array
                    array_length = self._get_array_length(target_variable_type, info.slot)

                    for j in range(min(array_length, self.max_depth)):
                        info = self.get_storage_slot(
                            var,
                            contract,
                            key=str(i),
                            deep_key=str(j),
                        )
                        if info:
                            elems[str(i)].elems[str(j)] = info
        return elems

    def _get_array_length(self, type_: Type, slot: int) -> int:
        """
        Gets the length of dynamic and fixed arrays.
        Args:
            type_ (`AbstractType`): The array type.
            slot (int): Slot a dynamic array's length is stored at.
        Returns:
            (int): The length of the array.
        """
        val = 0
        if self.rpc_info:
            # The length of dynamic arrays is stored at the starting slot.
            # Convert from hexadecimal to decimal.
            val = int(
                get_storage_data(
                    self.rpc_info.web3,
                    self.checksum_address,
                    int.to_bytes(slot, 32, byteorder="big"),
                    self.rpc_info.block,
                ).hex(),
                16,
            )
        if isinstance(type_, ArrayType):
            if type_.is_fixed_array:
                val = int(str(type_.length))

        return val



"""
Robert "Uncle Bob" Martin - Agile software metrics
https://en.wikipedia.org/wiki/Software_package_metrics

Efferent Coupling (Ce): Number of contracts that the contract depends on
Afferent Coupling (Ca): Number of contracts that depend on a contract
Instability (I): Ratio of efferent coupling to total coupling (Ce / (Ce + Ca))
Abstractness (A): Number of abstract contracts / total number of contracts
Distance from the Main Sequence (D):  abs(A + I - 1)

"""

from slither.printers.abstract_printer import AbstractPrinter
from slither.utils.martin import MartinMetrics
from slither.utils.output import Output


class Martin(AbstractPrinter):
    ARGUMENT = "martin"
    HELP = "Martin agile software metrics (Ca, Ce, I, A, D)"

    WIKI = "https://github.com/trailofbits/slither/wiki/Printer-documentation#martin"

    def output(self, _filename: str) -> Output:
        if len(self.contracts) == 0:
            return self.generate_output("No contract found")

        martin = MartinMetrics(self.contracts)

        res = self.generate_output(martin.full_text)
        res.add_pretty_table(martin.core.pretty_table, martin.core.title)
        self.info(martin.full_text)
        return res



"""
Structure module
"""

from typing import TYPE_CHECKING

from slither.core.compilation_unit import SlitherCompilationUnit
from slither.core.declarations.structure_top_level import StructureTopLevel
from slither.core.variables.structure_variable import StructureVariable
from slither.solc_parsing.declarations.caller_context import CallerContextExpression
from slither.solc_parsing.variables.structure_variable import StructureVariableSolc

if TYPE_CHECKING:
    from slither.solc_parsing.slither_compilation_unit_solc import SlitherCompilationUnitSolc


class StructureTopLevelSolc(CallerContextExpression):
    """
    Structure class
    """

    # elems = [(type, name)]

    def __init__(
        self,
        st: StructureTopLevel,
        struct: dict,
        slither_parser: "SlitherCompilationUnitSolc",
    ) -> None:
        if slither_parser.is_compact_ast:
            name = struct["name"]
            attributes = struct
        else:
            name = struct["attributes"][slither_parser.get_key()]
            attributes = struct["attributes"]
        if "canonicalName" in attributes:
            canonicalName = attributes["canonicalName"]
        else:
            canonicalName = name

        children = struct["members"] if "members" in struct else struct.get("children", [])

        self._structure = st
        st.name = name
        st.canonical_name = canonicalName
        self._slither_parser = slither_parser

        self._elemsNotParsed = children

    def analyze(self) -> None:
        for elem_to_parse in self._elemsNotParsed:
            elem = StructureVariable()
            elem.set_structure(self._structure)
            elem.set_offset(elem_to_parse["src"], self._slither_parser.compilation_unit)

            elem_parser = StructureVariableSolc(elem, elem_to_parse)
            elem_parser.analyze(self)

            self._structure.elems[elem.name] = elem
            self._structure.add_elem_in_order(elem.name)
        self._elemsNotParsed = []

    @property
    def is_compact_ast(self) -> bool:
        return self._slither_parser.is_compact_ast

    @property
    def compilation_unit(self) -> SlitherCompilationUnit:
        return self._slither_parser.compilation_unit

    def get_key(self) -> str:
        return self._slither_parser.get_key()

    @property
    def slither_parser(self) -> "SlitherCompilationUnitSolc":
        return self._slither_parser

    @property
    def underlying_structure(self) -> StructureTopLevel:
        return self._structure



from slither.core.declarations import Function, SolidityVariable
from slither.core.variables import Variable
from slither.slithir.operations.high_level_call import HighLevelCall
from slither.core.declarations.contract import Contract


class LibraryCall(HighLevelCall):
    """
    High level message call
    """

    # Development function, to be removed once the code is stable
    def _check_destination(self, destination: Variable | SolidityVariable | Contract) -> None:
        assert isinstance(destination, Contract)

    def can_reenter(self, callstack: list[Function | Variable] | None = None) -> bool:
        """
        Must be called after slithIR analysis pass
        :return: bool
        """
        if self.is_static_call():
            return False
        # In case of recursion, return False
        callstack_local = [] if callstack is None else callstack
        if self.function in callstack_local:
            return False
        callstack_local = callstack_local + [self.function]
        return self.function.can_reenter(callstack_local)

    def __str__(self):
        gas = ""
        if self.call_gas:
            gas = f"gas:{self.call_gas}"
        arguments = []
        if self.arguments:
            arguments = self.arguments
        if not self.lvalue:
            lvalue = ""
        elif isinstance(self.lvalue.type, (list,)):
            lvalue = f"{self.lvalue}({','.join(str(x) for x in self.lvalue.type)}) = "
        else:
            lvalue = f"{self.lvalue}({self.lvalue.type}) = "
        txt = "{}LIBRARY_CALL, dest:{}, function:{}, arguments:{} {}"

        function_name = self.function_name
        if self.function:
            if isinstance(self.function, Function):
                function_name = self.function.canonical_name
        return txt.format(
            lvalue,
            self.destination,
            function_name,
            [str(x) for x in arguments],
            gas,
        )



from slither.core.declarations import Contract
from slither.core.solidity_types.elementary_type import ElementaryType
from slither.core.solidity_types.type_alias import TypeAlias
from slither.core.solidity_types.user_defined_type import UserDefinedType
from slither.core.solidity_types.array_type import ArrayType
from slither.core.source_mapping.source_mapping import SourceMapping
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.utils.utils import is_valid_lvalue, is_valid_rvalue
from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA


class TypeConversion(OperationWithLValue):
    def __init__(
        self,
        result: TemporaryVariableSSA | TemporaryVariable,
        variable: SourceMapping,
        variable_type: TypeAlias | UserDefinedType | ElementaryType,
    ) -> None:
        super().__init__()
        assert is_valid_rvalue(variable) or isinstance(variable, Contract)
        assert is_valid_lvalue(result)
        assert isinstance(variable_type, (TypeAlias, UserDefinedType, ElementaryType, ArrayType))

        self._variable = variable
        self._type: TypeAlias | UserDefinedType | ElementaryType | ArrayType = variable_type
        self._lvalue = result

    @property
    def variable(self) -> SourceMapping:
        return self._variable

    @property
    def type(
        self,
    ) -> TypeAlias | UserDefinedType | ElementaryType:
        return self._type

    @property
    def read(self) -> list[SourceMapping]:
        return [self.variable]

    def __str__(self) -> str:
        return str(self.lvalue) + f" = CONVERT {self.variable} to {self.type}"



from typing import Any
from slither.core.declarations import Modifier
from slither.core.declarations.function import Function
from slither.core.declarations.function_contract import FunctionContract
from slither.slithir.operations.call import Call
from slither.slithir.operations.lvalue import OperationWithLValue
from slither.slithir.variables.constant import Constant
from slither.slithir.variables.temporary import TemporaryVariable
from slither.slithir.variables.temporary_ssa import TemporaryVariableSSA
from slither.slithir.variables.tuple import TupleVariable
from slither.slithir.variables.tuple_ssa import TupleVariableSSA


class InternalCall(Call, OperationWithLValue):
    def __init__(
        self,
        function: Function | tuple[str, str],
        nbr_arguments: int,
        result: TupleVariableSSA | TemporaryVariableSSA | TupleVariable | TemporaryVariable | None,
        type_call: str,
        names: list[str] | None = None,
    ) -> None:
        """
        #### Parameters
        names -
            For calls of the form f({argName1 : arg1, ...}), the names of parameters listed in call order.
            Otherwise, None.
        """
        super().__init__(names=names)
        self._contract_name = ""
        if isinstance(function, Function):
            self._function: Function | None = function
            self._function_name = function.name
            if isinstance(function, FunctionContract):
                self._contract_name = function.contract_declarer.name
        else:
            self._function = None
            self._function_name, self._contract_name = function
        # self._contract = contract
        self._nbr_arguments = nbr_arguments
        self._type_call = type_call
        self._lvalue = result
        # function_candidates is only used as an helper to retrieve the "function" object
        # For top level function called through a import renamed
        # See SolidityImportPlaceHolder usages
        self.function_candidates: list[Function] | None = None

    @property
    def read(self) -> list[Any]:
        return list(self._unroll(self.arguments))

    @property
    def function(self) -> Function | None:
        return self._function

    @function.setter
    def function(self, f):
        self._function = f

    @property
    def function_name(self) -> Constant:
        return self._function_name

    @property
    def contract_name(self) -> str:
        return self._contract_name

    @property
    def nbr_arguments(self) -> int:
        return self._nbr_arguments

    @property
    def type_call(self) -> str:
        return self._type_call

    @property
    def is_modifier_call(self):
        """
        Check if the destination is a modifier
        :return: bool
        """
        return isinstance(self.function, Modifier)

    def __str__(self):
        args = [str(a) for a in self.arguments]
        if not self.lvalue:
            lvalue = ""
        elif isinstance(self.lvalue.type, (list,)):
            lvalue = f"{self.lvalue}({','.join(str(x) for x in self.lvalue.type)}) = "
        else:
            lvalue = f"{self.lvalue}({self.lvalue.type}) = "
        if self.is_modifier_call:
            txt = "{}MODIFIER_CALL, {}({})"
        else:
            txt = "{}INTERNAL_CALL, {}({})"
        return txt.format(lvalue, self.function.canonical_name, ",".join(args))
